/******************************************************************************
 *
 * Name   : sky2.c
 * Project: Gigabit Ethernet Driver for FreeBSD 5.x/6.x
 * Version: $Revision: 1.23 $
 * Date   : $Date: 2005/12/22 09:04:11 $
 * Purpose: Main driver source file
 *
 *****************************************************************************/

/******************************************************************************
 *
 *	LICENSE:
 *	Copyright (C) Marvell International Ltd. and/or its affiliates
 *
 *	The computer program files contained in this folder ("Files")
 *	are provided to you under the BSD-type license terms provided
 *	below, and any use of such Files and any derivative works
 *	thereof created by you shall be governed by the following terms
 *	and conditions:
 *
 *	- Redistributions of source code must retain the above copyright
 *	  notice, this list of conditions and the following disclaimer.
 *	- Redistributions in binary form must reproduce the above
 *	  copyright notice, this list of conditions and the following
 *	  disclaimer in the documentation and/or other materials provided
 *	  with the distribution.
 *	- Neither the name of Marvell nor the names of its contributors
 *	  may be used to endorse or promote products derived from this
 *	  software without specific prior written permission.
 *
 *	THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *	"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *	LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 *	FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 *	COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 *	INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 *	BUT NOT LIMITED TO, PROCUREMENT OF  SUBSTITUTE GOODS OR SERVICES;
 *	LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 *	HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 *	STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *	ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 *	OF THE POSSIBILITY OF SUCH DAMAGE.
 *	/LICENSE
 *
 *****************************************************************************/

/*-
 * Copyright (c) 1997, 1998, 1999, 2000
 *	Bill Paul <wpaul@ctr.columbia.edu>.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by Bill Paul.
 * 4. Neither the name of the author nor the names of any co-contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY Bill Paul AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL Bill Paul OR THE VOICES IN HIS HEAD
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */
/*-
 * Copyright (c) 2003 Nathan L. Binkert <binkertn@umich.edu>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

/*
 * Device driver for the Marvell Yukon II Ethernet controller.
 * Due to lack of documentation, this driver is based on the code from
 * sk(4) and Marvell's myk(4) driver for FreeBSD 5.x.
 */

/*
 Change log
 */

/*
 TODO:
	port reset and global reset.
 */

/*
 * System Header files.
 */
#include <sys/types.h>
#include <sys/conf.h>
#include <sys/debug.h>
#include <sys/stropts.h>
#include <sys/stream.h>
#include <sys/strlog.h>
#include <sys/kmem.h>
#include <sys/stat.h>
#include <sys/kstat.h>
#include <sys/vtrace.h>
#include <sys/dlpi.h>
#include <sys/strsun.h>
#include <sys/ethernet.h>
#include <sys/modctl.h>
#include <sys/errno.h>
#include <sys/dditypes.h>
#include <sys/ddi.h>
#include <sys/sunddi.h>

#include <sys/pci.h>
#include "mii.h"
#include "gem.h"

#include "mykreg.h"

#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>

char	ident[] = "yukon2 GbE driver v" VERSION;

/* Debugging support */
#ifdef DEBUG_LEVEL
static int myk_debug = DEBUG_LEVEL;
#define	DPRINTF(n, args)	if (myk_debug>(n)) cmn_err args
#else
#define	DPRINTF(n, args)
#endif

/*
 * Useful macros and typedefs
 */
#ifndef PCI_CAP_ID_PCI_E
#  define	PCI_CAP_ID_PCI_E	0x10
#endif
#ifndef	PCI_PMCSR_D3COLD
#  define	PCI_PMCSR_D3COLD	4
#endif

#if defined(i86pc)
#define	LE16(x)	((uint16_t) (x))
#define	LE32(x)	((uint32_t) (x))
#define	LE64(x)	((uint64_t) (x))
#else
#define	LE16(x)	ddi_swap16(x)
#define	LE32(x)	ddi_swap32(x)
#define	LE64(x)	ddi_swap64(x)
#endif

#define	FLSHB(dp, reg)	(void)INB(dp, reg)
#define	FLSHW(dp, reg)	(void)INW(dp, reg)
#define	FLSHL(dp, reg)	(void)INL(dp, reg)

#define	ROUNDUP2(x, y)	(((x)+(y)-1) & ~((y)-1))

#define	INSIDE(slot, head, tail)	\
	(((head) <= (tail)) ?	\
		((head) <= (slot) && (slot) < (tail)) :	\
		((slot) < (tail)  || (head) <= (slot)))
#define	offsetof(t, m)	((long) &(((t *) 0)->m))

#define	prefetch()

#define	RAM_BUFFER(b, r)	SELECT_RAM_BUFFER(b, r)
#define	Y2_QADDR(q, r)		Y2_PREF_Q_ADDR(q, r)

/*
 * Our configuration
 */
#ifndef TX_RING_SIZE
# define TX_RING_SIZE	256
#endif
#ifndef TX_BUF_SIZE
# define TX_BUF_SIZE	TX_RING_SIZE
#endif

#ifndef RX_RING_SIZE
# define RX_RING_SIZE	256
#endif
#ifndef RX_BUF_SIZE
# define RX_BUF_SIZE	RX_RING_SIZE
#endif

#define	DESC_ALIGN	0x1000
#define	ONESEC			(drv_usectohz(1*1000000))

/* I/O area confiuration */

#define STATUS_RING_SIZE	2048	/* 2 ports * (TX + 2*RX) */
#define STATUS_LE_BYTES		\
	ROUNDUP2(STATUS_RING_SIZE * sizeof(struct msk_stat_desc), DESC_ALIGN)
#define	RX_RING_BYTES	\
	ROUNDUP2(RX_RING_SIZE * sizeof(struct msk_rx_desc), DESC_ALIGN)

#define	RX_RING_BASE(dp)		((dp)->io_area)
#define	STATUS_RING_BASE(dp)		((dp)->io_area + RX_RING_BYTES)

#define	RX_RING_BASE_DMA(dp)		((dp)->io_area_dma)
#define	STATUS_RING_BASE_DMA(dp)	((dp)->io_area_dma + RX_RING_BYTES)

/*
 * Marvell GbE diriver driver for Solaris
 */
struct rx_soft_desc {
	uint32_t	rx_len;
	uint16_t	rx_stat;
	uint16_t	rx_tag;
	uint16_t	rx_cksum;
};

#define	PORT(dp)	0

struct myk_dev {
	kmutex_t	chip_lock;

	uint8_t		mac_id;
	uint8_t		chip_id;
	uint8_t		chip_rev;
	uint8_t		ports;
	uint8_t		pmd_type;
	uint_t		pm_cap;
	uint_t		is_pcie;

	ddi_iblock_cookie_t   iblock_cookie;

	seqnum_t	status_desc_tail;

	struct myk_port {
		kmutex_t	port_lock;

		struct gem_dev	*dp;

		uint32_t	imsk;
		uint32_t	hwe_imsk;
		uint32_t	isr;

		uint32_t	tx_hiaddr;
		uint_t		tx_head;
		uint_t		tx_tail;

		uint32_t	rx_hiaddr;
		seqnum_t	rx_desc_head;
		seqnum_t	rx_desc_tail;
		seqnum_t	rx_slot_head;
		seqnum_t	rx_slot_tail;
		struct rx_soft_desc	rx_slot[RX_RING_SIZE];
	} descset[2];
};

static int	myk_tx_copy_thresh = 256;
static int	myk_rx_copy_thresh = 256;

/* ======================================================== */
 
/* mii operations */
static void  myk_mii_sync(struct gem_dev *);
static uint16_t myk_mii_read(struct gem_dev *, uint_t);
static void myk_mii_write(struct gem_dev *, uint_t, uint16_t);

/* nic operations */
static int myk_attach_port(struct gem_dev *);
static int myk_reset_port(struct gem_dev *);
static int myk_init_port(struct gem_dev *);
static int myk_start_port(struct gem_dev *);
static int myk_stop_port(struct gem_dev *);
static int myk_set_media(struct gem_dev *);
static int myk_set_rx_filter(struct gem_dev *);
static int myk_get_stats(struct gem_dev *);

/* descriptor operations */
static int myk_tx_desc_write(struct gem_dev *dp, int slot,
		    ddi_dma_cookie_t *dmacookie, int frags, uint32_t flag);
static void myk_tx_start(struct gem_dev *dp, int slot, int nslot);
static void myk_rx_start(struct gem_dev *dp, int slot, int nslot);
static void myk_rx_desc_write(struct gem_dev *dp, int slot,
		    ddi_dma_cookie_t *dmacookie, int frags);
static uint_t myk_tx_desc_stat(struct gem_dev *dp, int slot, int ndesc);
static uint64_t myk_rx_desc_stat(struct gem_dev *dp, int slot, int ndesc);

static void myk_tx_desc_init(struct gem_dev *dp, int slot);
static void myk_rx_desc_init(struct gem_dev *dp, int slot);
static void myk_tx_desc_clean(struct gem_dev *dp, int slot);
static void myk_rx_desc_clean(struct gem_dev *dp, int slot);

/* interrupt handler */
static u_int myk_soft_intr(struct gem_dev *dp);

/* ======================================================== */

/* mapping attributes */
/* Data access requirements. */
static struct ddi_device_acc_attr myk_dev_attr = {
	DDI_DEVICE_ATTR_V0,
	DDI_STRUCTURE_LE_ACC,
	DDI_STRICTORDER_ACC
};

/* On sparc, the buffers should be native endianness */
static struct ddi_device_acc_attr myk_buf_attr = {
	DDI_DEVICE_ATTR_V0,
	DDI_NEVERSWAP_ACC,	/* native endianness */
	DDI_STRICTORDER_ACC
};

static ddi_dma_attr_t myk_dma_attr_buf = {
	DMA_ATTR_V0,		/* dma_attr_version */
	0,			/* dma_attr_addr_lo */
	0x000000ffffffffffull,	/* dma_attr_addr_hi */
	0x000000000000ffffull,	/* dma_attr_count_max */
	1,			/* dma_attr_align */
	0x000007fc,		/* dma_attr_burstsizes */
	1,			/* dma_attr_minxfer */
	0x000000000000ffffull,	/* dma_attr_maxxfer */
	0x000000ffffffffffull,	/* dma_attr_seg */
	0,/* patched later */	/* dma_attr_sgllen */
	1,			/* dma_attr_granular */
	0			/* dma_attr_flags */
};

static ddi_dma_attr_t myk_dma_attr_desc = {
	DMA_ATTR_V0,		/* dma_attr_version */
	0,			/* dma_attr_addr_lo */
	0x000000ffffffffffull,	/* dma_attr_addr_hi */
	0x00000000ffffffffull,	/* dma_attr_count_max */
	DESC_ALIGN,		/* dma_attr_align */
	0x000007fc,		/* dma_attr_burstsizes */
	1,			/* dma_attr_minxfer */
	0x00000000ffffffffull,	/* dma_attr_maxxfer */
	0x000000ffffffffffull,	/* dma_attr_seg */
	1,			/* dma_attr_sgllen */
	1,			/* dma_attr_granular */
	0			/* dma_attr_flags */
};

#include "mskreg.h"

static int
myk_is_copper(struct gem_dev *dp)
{
	boolean_t	ret;
	struct myk_dev	*lp = (struct myk_dev *) dp->private;

	switch (lp->pmd_type) {
	case 'L': case 'S': case 'P':
		ret = B_FALSE;
		break;

	default:
		ret = B_TRUE;
		break;
	}

	return (ret);
}

/* Yukon PHY related registers */
#define GMAC_REG(port,reg) \
	(BASE_GMAC_1 + (port) * (BASE_GMAC_2-BASE_GMAC_1) + (reg))

static uint16_t
gma_read16(struct gem_dev *dp, unsigned port, unsigned reg)
{
	return (INW(dp, GMAC_REG(port, reg)));
}

static uint32_t
gma_read32(struct gem_dev *dp, unsigned port, unsigned reg)
{
	unsigned base = GMAC_REG(port, reg);

	return ((uint32_t) (INW(dp, base) | (INW(dp, base + 4) << 16)));
}

static void
gma_write16(struct gem_dev *dp, unsigned port, int r, uint16_t v)
{
	OUTW(dp, GMAC_REG(port, r), v);
}

static void
gma_set_addr(struct gem_dev *dp, unsigned port, unsigned reg, uint8_t *addr)
{
	OUTW(dp, GMAC_REG(port, reg + 0), addr[0] | (addr[1] << 8));
	OUTW(dp, GMAC_REG(port, reg + 4), addr[2] | (addr[3] << 8));
	OUTW(dp, GMAC_REG(port, reg + 8), addr[4] | (addr[5] << 8));
}

/* PCI config space access */
static uint32_t
myk_pci_read32(struct gem_dev *dp, unsigned reg)
{
	return (INL(dp, Y2_CFG_SPC + reg));
}

static uint16_t
myk_pci_read16(struct gem_dev *dp, unsigned reg)
{
	return (INW(dp, Y2_CFG_SPC + reg));
}

static void
myk_pci_write32(struct gem_dev *dp, unsigned reg, uint32_t val)
{
	OUTL(dp, Y2_CFG_SPC + reg, val);
}

static void
myk_pci_write16(struct gem_dev *dp, unsigned reg, uint16_t val)
{
	OUTW(dp, Y2_CFG_SPC + reg, val);
}

/*
 * The Yukon II chipset takes 64 bit command blocks (called list elements)
 * that are organized into three (receive, transmit, status) different rings
 * similar to Tigon3. A transmit can require several elements;
 * a receive requires one (or two if using 64 bit dma).
 */
static int copybreak = 256;
static int disable_msi = 0;
static int idle_timeout = 100;

/* Avoid conditionals by using array */
#define TXQADDR(p)	((p) == 0 ? Q_XA1 : Q_XA2)
#define RXQADDR(p)	((p) == 0 ? Q_R1 : Q_R2)
#define PORTIRQ_MSK(p)	((p) == 0 ? Y2_IS_PORT_A : Y2_IS_PORT_B)

/* This driver supports yukon2 chipset only */
static const char *yukon2_name[] = {
	"XL",		/* 0xb3 */
	"EC Ultra", 	/* 0xb4 */
	"UNKNOWN",	/* 0xb5 */
	"EC",		/* 0xb6 */
	"FE",		/* 0xb7 */
};

static void
myk_set_power_state(struct gem_dev *dp, unsigned state)
{
	uint16_t	power_control;
	uint32_t	val;
	int		i;
	int		vaux;
	struct myk_dev	*lp = (struct myk_dev *) dp->private;

	DPRINTF(0, (CE_CONT, "%s %d", dp->name, state));

	OUTB(dp, B2_TST_CTRL1, TST_CFG_WRITE_ON);

	power_control = myk_pci_read16(dp, lp->pm_cap + PCI_PMCAP);
	vaux = (INW(dp, B0_CTST) & Y2_VAUX_AVAIL) &&
		(power_control & PCI_PMCAP_D3COLD_PME);

	power_control = myk_pci_read16(dp, lp->pm_cap + PCI_PMCSR);

	power_control |= PCI_PMCSR_PME_STAT;
	power_control &= ~(PCI_PMCSR_STATE_MASK);

#define	CHIP_REV_YU_XL_A1	1

	switch (state) {
	case PCI_PMCSR_D0:
		/* switch power to VCC (WA for VAUX problem) */
		OUTB(dp, B0_POWER_CTRL,
			PC_VAUX_ENA | PC_VCC_ENA | PC_VAUX_OFF | PC_VCC_ON);

		/* disable Core Clock Division, */
		OUTL(dp, B2_Y2_CLK_CTRL, Y2_CLK_DIV_DIS);

		val = 0;
		if (lp->chip_id == CHIP_ID_YUKON_XL &&
		    lp->chip_rev > CHIP_REV_YU_XL_A1) {
			/* Enable bits are inverted */
			val = Y2_PCI_CLK_LNK1_DIS | Y2_COR_CLK_LNK1_DIS |
			      Y2_CLK_GAT_LNK1_DIS | Y2_PCI_CLK_LNK2_DIS |
			      Y2_COR_CLK_LNK2_DIS | Y2_CLK_GAT_LNK2_DIS;
		}
		/*
		 * Enable PCI& Core Clock, enable clock gating for both Links.
		 */
		OUTB(dp, B2_Y2_CLK_GATE, val);

		/* Turn off phy power saving */
		val = myk_pci_read32(dp, PCI_OUR_REG_1);
		val &= ~(PCI_Y2_PHY1_POWD | PCI_Y2_PHY2_POWD);

		/* looks like this XL is back asswards .. */
		if (lp->chip_id == CHIP_ID_YUKON_XL &&
		    lp->chip_rev > CHIP_REV_YU_XL_A1) {
			val |= PCI_Y2_PHY1_COMA;
			if (lp->ports > 1) {
				val |= PCI_Y2_PHY2_COMA;
			}
		} else if (lp->chip_id == CHIP_ID_YUKON_EC_U) {
			uint32_t	our;

			OUTW(dp, B0_CTST, Y2_HW_WOL_ON);

			/* Enable all clocks. */
			myk_pci_write32(dp, PCI_OUR_REG_3, 0);
			our = myk_pci_read32(dp, PCI_OUR_REG_4);
			our &= (PCI_FORCE_ASPM_REQUEST|PCI_ASPM_GPHY_LINK_DOWN|
			    PCI_ASPM_INT_FIFO_EMPTY|PCI_ASPM_CLKRUN_REQUEST);
			/* Set all bits to 0 except bits 15..12. */
			myk_pci_write32(dp, PCI_OUR_REG_4, our);
			/* Set to default value. */
			myk_pci_write32(dp, PCI_OUR_REG_5, 0);
		}
		/* Release PHY from PowerDown/COMA mode. */
		myk_pci_write32(dp, PCI_OUR_REG_1, val);
		for (i = 0; i < lp->ports; i++) {
			OUTW(dp, MR_ADDR(i, GMAC_LINK_CTRL), GMLC_RST_SET);
			OUTW(dp, MR_ADDR(i, GMAC_LINK_CTRL), GMLC_RST_CLR);
		}
		break;

	case PCI_PMCSR_D3HOT:
	case PCI_PMCSR_D3COLD:
		/* Turn on phy power saving */
		val = myk_pci_read32(dp, PCI_OUR_REG_1);
		if (lp->chip_id == CHIP_ID_YUKON_XL && lp->chip_rev > 1) {
			val &= ~(PCI_Y2_PHY1_POWD | PCI_Y2_PHY2_POWD);
		} else {
			val |= (PCI_Y2_PHY1_POWD | PCI_Y2_PHY2_POWD);
		}
		myk_pci_write32(dp, PCI_OUR_REG_1, val);

		if (lp->chip_id == CHIP_ID_YUKON_XL && lp->chip_rev > 1) {
			OUTB(dp, B2_Y2_CLK_GATE, 0);
		} else {
			/* enable bits are inverted */
			OUTB(dp, B2_Y2_CLK_GATE,
				    Y2_PCI_CLK_LNK1_DIS | Y2_COR_CLK_LNK1_DIS |
				    Y2_CLK_GAT_LNK1_DIS | Y2_PCI_CLK_LNK2_DIS |
				    Y2_COR_CLK_LNK2_DIS | Y2_CLK_GAT_LNK2_DIS);
		}

		/* switch power to VAUX */
		if (vaux && state != PCI_PMCSR_D3COLD) {
			OUTB(dp, B0_POWER_CTRL,
				    (PC_VAUX_ENA | PC_VCC_ENA |
				     PC_VAUX_ON | PC_VCC_OFF));
		}
		break;

	default:
		cmn_err(CE_WARN, "%s: %s; Unknown power state %d",
		dp->name, __func__, state);
		break;
	}

	myk_pci_write16(dp, lp->pm_cap + PCI_PMCSR, power_control);
	OUTB(dp, B2_TST_CTRL1, TST_CFG_WRITE_OFF);
}

/* Chip internal frequency for clock calculations */
static uint32_t
myk_mhz(struct gem_dev *dp)
{
	struct myk_dev	*lp = (struct myk_dev *) dp->private;

	switch (lp->chip_id) {
	case CHIP_ID_YUKON_EC:
	case CHIP_ID_YUKON_EC_U:
		return (125);	/* 125 Mhz */
	case CHIP_ID_YUKON_FE:
		return (100);	/* 100 Mhz */
	default:		/* YUKON_XL */
		return (156);	/* 156 Mhz */
	}
}

static uint32_t
myk_us2clk(struct gem_dev *dp, uint32_t us)
{
	return (myk_mhz(dp) * us);
}

static uint32_t
myk_clk2us(struct gem_dev *dp, uint32_t clk)
{
	return (clk / myk_mhz(dp));
}

/* ======================================================== */
/*
 * status descriptor setup  routines
 */
/* ======================================================== */
static int
myk_set_coalesce(struct gem_dev *dp)
{
	uint32_t	tmax;
	uint32_t	tx_coalesce_usecs = 500*1000;
	uint16_t	tx_max_coalesced_frames = TX_RING_SIZE/2;
	uint16_t	rx_coalesce_usecs = 0;
	uint8_t		rx_max_coalesced_frames = dp->poll_pkt_delay;
	uint32_t	rx_coalesce_usecs_irq = 0;
	uint8_t		rx_max_coalesced_frames_irq = dp->poll_pkt_delay;

	tmax = myk_clk2us(dp, 0x0ffffff);
#ifdef notdef
	if (ecmd->tx_max_coalesced_frames >= TX_RING_SIZE-1)
		return -EINVAL;
	if (ecmd->rx_max_coalesced_frames > RX_MAX_PENDING)
		return -EINVAL;
	if (ecmd->rx_max_coalesced_frames_irq >RX_MAX_PENDING)
		return -EINVAL;
#endif
	if (tx_coalesce_usecs == 0) {
		OUTB(dp, STAT_TX_TIMER_CTRL, TIM_STOP);
	} else {
		OUTL(dp, STAT_TX_TIMER_INI,
			     myk_us2clk(dp, tx_coalesce_usecs));
		OUTB(dp, STAT_TX_TIMER_CTRL, TIM_START);
	}
	OUTW(dp, STAT_TX_IDX_TH, tx_max_coalesced_frames);

	if (rx_coalesce_usecs == 0) {
		OUTB(dp, STAT_LEV_TIMER_CTRL, TIM_STOP);
	} else {
		OUTL(dp, STAT_LEV_TIMER_INI,
			     myk_us2clk(dp, rx_coalesce_usecs));
		OUTB(dp, STAT_LEV_TIMER_CTRL, TIM_START);
	}
	OUTB(dp, STAT_FIFO_WM, rx_max_coalesced_frames);

	if (rx_coalesce_usecs_irq == 0) {
		OUTB(dp, STAT_ISR_TIMER_CTRL, TIM_STOP);
	} else {
		OUTL(dp, STAT_ISR_TIMER_INI,
			     myk_us2clk(dp, rx_coalesce_usecs_irq));
		OUTB(dp, STAT_ISR_TIMER_CTRL, TIM_START);
	}
	OUTB(dp, STAT_FIFO_ISR_WM, rx_max_coalesced_frames_irq);

	return 0;
}

/* ======================================================== */
/*
 * HW descriputor cache sync routines
 */
/* ======================================================== */
static void
myk_rx_desc_dma_sync(struct gem_dev *dp, int head, int nslot, int how)
{
	int	n;
	int	m;
	off_t	rx_desc_offset;

	/* sync active descriptors */
	rx_desc_offset = (off_t) (RX_RING_BASE_DMA(dp) - dp->rx_ring_dma),
	n = RX_RING_SIZE - head;
	if ((m = nslot - n) > 0) {
		ddi_dma_sync(dp->desc_dma_handle, rx_desc_offset,
			(size_t) (m * sizeof(struct msk_rx_desc)),
			 how);
		nslot = n;
	}

	(void) ddi_dma_sync(dp->desc_dma_handle,
		(off_t)(head * sizeof(struct msk_rx_desc)
			+ rx_desc_offset),
		(size_t) (nslot * sizeof(struct msk_rx_desc)),
		how);
}

static void
myk_status_desc_dma_sync(struct gem_dev *dp, int head, int nslot, int how)
{
	int	n;
	int	m;
	off_t	status_desc_offset;

	/* sync active descriptors */
	status_desc_offset = STATUS_RING_BASE_DMA(dp) - dp->rx_ring_dma,
	n = STATUS_RING_SIZE - head;
	if ((m = nslot - n) > 0) {
		ddi_dma_sync(dp->desc_dma_handle, status_desc_offset,
			(size_t) (m * sizeof(struct msk_stat_desc)),
			 how);
		nslot = n;
	}

	(void) ddi_dma_sync(dp->desc_dma_handle,
		(off_t)(head * sizeof(struct msk_stat_desc)
			+ status_desc_offset),
		(size_t) (nslot * sizeof(struct msk_stat_desc)),
		how);
}

/* ======================================================== */
/*
 * HW manupilation routines
 */
/* ======================================================== */
static boolean_t
myk_test_chip(struct gem_dev *dp, char *msg)
{
	uint32_t	val;
	boolean_t	ret;

#define	TEST_VAL	0x11335577
#define	B2_IRQM_INI	0x0140

	/* check if the adapter seems to be accessible */
	val = INL(dp, B2_IRQM_INI);
	OUTL(dp, B2_IRQM_INI, TEST_VAL);

	if (ret = (INL(dp, B2_IRQM_INI) == TEST_VAL)) {
		/* restore original value */
		cmn_err(CE_CONT, "%s: test succeeded for %s",
			dp->name, msg);
		OUTL(dp, B2_IRQM_INI, val);
	} else {
		cmn_err(CE_CONT, "%s: test failed for %s",
			dp->name, msg);
		OUTL(dp, B2_IRQM_INI, 0);
	}

	return (ret);
}

static int
myk_reset_port(struct gem_dev *dp)
{
	int		i;
	u_int		port = PORT(dp);
	static int	is_initialized = B_FALSE;
	struct myk_dev	*lp = (struct myk_dev *) dp->private;

	DPRINTF(0, (CE_CONT, "!%s: %s: called, GPHY:%08x, GMAC:%08x",
		dp->name, __func__,
		INL(dp, MR_ADDR(port, GPHY_CTRL)),
		INL(dp, MR_ADDR(port, GMAC_CTRL))));

	/* setup back pointer to struct gem_dev */
	if (lp->descset[port].dp == NULL) {
		lp->descset[port].dp = dp;
	}

	if (!is_initialized) {
#ifdef notdef	/* shoud be moved into attach */
		myk_test_chip(dp, "reset_port@1");
#endif /* notdef */

		/*
		 * Since we disabled ASF, S/W reset is required
		 * for Power Management.
		 */
		OUTW(dp, B0_CTST, CS_RST_SET);
		OUTW(dp, B0_CTST, CS_RST_CLR);

		/* make config registers writable */
		OUTB(dp, B2_TST_CTRL1, TST_CFG_WRITE_ON);

		/* Clear all error bits in the PCI status register */
		myk_pci_write16(dp, PCI_CONF_STAT,
		    myk_pci_read16(dp, PCI_CONF_STAT) |
		    PCI_STAT_PERROR | PCI_STAT_S_SYSERR | PCI_STAT_R_MAST_AB |
		    PCI_STAT_R_TARG_AB | PCI_STAT_S_PERROR);

		OUTW(dp, B0_CTST, CS_MRST_CLR);

		if (lp->is_pcie) {
			/* clear any PEX errors */
			myk_pci_write32(dp, PEX_UNC_ERR_STAT, ~0u);
		}
		/* case PCI */
		/* case PCIX */

		/* Set PHY power state. */
		myk_set_power_state(dp, PCI_PMCSR_D0);

		/* make config registers read only */
		OUTB(dp, B2_TST_CTRL1, TST_CFG_WRITE_OFF);

		/* LED On. */
		OUTW(dp, B0_CTST, Y2_LED_STAT_ON);

		/* Clear TWSI IRQ. */
		OUTL(dp, B2_I2C_IRQ, 1/* I2C_CLR_IR*/);

		/* Turn off hardware timer. */
		OUTB(dp, B2_TI_CTRL, TIM_STOP);
		OUTB(dp, B2_TI_CTRL, TIM_CLR_IRQ);

		/* Turn off descriptor polling. */
		OUTB(dp, B28_DPT_CTRL, DPT_STOP);

		/* Turn off time stamps. */
		OUTB(dp, GMAC_TI_ST_CTRL, GMT_ST_STOP);
		OUTB(dp, GMAC_TI_ST_CTRL, GMT_ST_CLR_IRQ);

		/* inhibit all interrupts */
		OUTL(dp, B0_HWE_IMSK, 0);
		FLSHL(dp, B0_HWE_IMSK);
		OUTL(dp, B0_IMSK, 0);
		FLSHL(dp, B0_IMSK);

		for (i = 0; i < lp->ports; i++) {
			lp->descset[i].imsk = 0;
			lp->descset[i].hwe_imsk = 0;
		}

		/* XXX - need for workaround for PCI-X cards */
		/* XXX - need for workaround for PCIE cards */
		if (lp->is_pcie) {
			uint16_t v;

			v = myk_pci_read16(dp, PEX_DEV_CTRL);
			/* Change Max. Read Request Size to 4096 bytes. */
			v &= ~PEX_DC_MAX_RRS_MSK;
			v |= PEX_DC_MAX_RD_RQ_SIZE(5);
			myk_pci_write16(dp, PEX_DEV_CTRL, v);
		}
#ifdef sun4u
		/* byte swap descriptors in hardware */
		myk_pci_write32(dp, PCI_OUR_REG_2,
			myk_pci_read32(dp, PCI_OUR_REG_2) | PCI_REV_DESC);
#endif
		is_initialized = B_TRUE;
	}

	/* inhibit all interrupts */
	lp->descset[port].hwe_imsk = 0;
	OUTL(dp, B0_HWE_IMSK,
		lp->descset[0].hwe_imsk | lp->descset[1].hwe_imsk);
	FLSHL(dp, B0_HWE_IMSK);

	lp->descset[port].imsk = 0;
	OUTL(dp, B0_IMSK, lp->descset[0].imsk | lp->descset[1].imsk);
	FLSHL(dp, B0_IMSK);

	/* Reset GPHY/GMAC Control */
	/* GPHY Control reset. */
	OUTL(dp, MR_ADDR(port, GPHY_CTRL), GPC_RST_SET);
	OUTL(dp, MR_ADDR(port, GPHY_CTRL), GPC_RST_CLR);

	/* GMAC Control reset. */
	OUTL(dp, MR_ADDR(port, GMAC_CTRL), GMC_RST_SET);
	OUTL(dp, MR_ADDR(port, GMAC_CTRL), GMC_RST_CLR);
	OUTL(dp, MR_ADDR(port, GMAC_CTRL), GMC_F_LOOPB_OFF);

	/* Configure timeout values. */
	OUTW(dp, RAM_BUFFER(port, B3_RI_CTRL), RI_RST_SET);
	OUTW(dp, RAM_BUFFER(port, B3_RI_CTRL), RI_RST_CLR);

	OUTB(dp, RAM_BUFFER(port, B3_RI_WTO_R1), MSK_RI_TO_53);
	OUTB(dp, RAM_BUFFER(port, B3_RI_WTO_XA1), MSK_RI_TO_53);
	OUTB(dp, RAM_BUFFER(port, B3_RI_WTO_XS1), MSK_RI_TO_53);
	OUTB(dp, RAM_BUFFER(port, B3_RI_RTO_R1), MSK_RI_TO_53);
	OUTB(dp, RAM_BUFFER(port, B3_RI_RTO_XA1), MSK_RI_TO_53);
	OUTB(dp, RAM_BUFFER(port, B3_RI_RTO_XS1), MSK_RI_TO_53);
	OUTB(dp, RAM_BUFFER(port, B3_RI_WTO_R2), MSK_RI_TO_53);
	OUTB(dp, RAM_BUFFER(port, B3_RI_WTO_XA2), MSK_RI_TO_53);
	OUTB(dp, RAM_BUFFER(port, B3_RI_WTO_XS2), MSK_RI_TO_53);
	OUTB(dp, RAM_BUFFER(port, B3_RI_RTO_R2), MSK_RI_TO_53);
	OUTB(dp, RAM_BUFFER(port, B3_RI_RTO_XA2), MSK_RI_TO_53);
	OUTB(dp, RAM_BUFFER(port, B3_RI_RTO_XS2), MSK_RI_TO_53);

	DPRINTF(0, (CE_CONT, "!%s: %s: exit, GPHY:%08x, GMAC:%08x",
		dp->name, __func__,
		INL(dp, MR_ADDR(port, GPHY_CTRL)),
		INL(dp, MR_ADDR(port, GMAC_CTRL))));

	return (GEM_SUCCESS);
}

/*
 * Assign Ram Buffer allocation.
 * start and end are in units of 4k bytes
 * ram registers are in units of 64bit words
 */
static void
myk_rambuffer_init(struct gem_dev *dp, uint16_t q, uint8_t startk, uint8_t endk)
{
	uint32_t	start;
	uint32_t	end;

	start = startk * 4096/8;
	end = (endk * 4096/8) - 1;

	OUTB(dp, RB_ADDR(q, RB_CTRL), RB_RST_CLR);
	OUTL(dp, RB_ADDR(q, RB_START), start);
	OUTL(dp, RB_ADDR(q, RB_END), end);
	OUTL(dp, RB_ADDR(q, RB_WP), start);
	OUTL(dp, RB_ADDR(q, RB_RP), start);

	if (q == Q_R1 || q == Q_R2) {
		uint32_t	space = (endk - startk) * 4096/8;
		uint32_t	tp = space - space/4;

		/*
		 * On receive queue's set the thresholds
		 * give receiver priority when > 3/4 full
		 * send pause when down to 2K
		 */
		OUTL(dp, RB_ADDR(q, RB_RX_UTHP), tp);
		OUTL(dp, RB_ADDR(q, RB_RX_LTHP), space/2);

		tp = space - 2048/8;
		OUTL(dp, RB_ADDR(q, RB_RX_UTPP), tp);
		OUTL(dp, RB_ADDR(q, RB_RX_LTPP), space/4);
	} else {
		/*
		 * Enable store & forward on Tx queue's because
		 * Tx FIFO is only 1K on Yukon
		 */
		OUTB(dp, RB_ADDR(q, RB_CTRL), RB_ENA_STFWD);
	}

	OUTB(dp, RB_ADDR(q, RB_CTRL), RB_ENA_OP_MD);
	FLSHB(dp, RB_ADDR(q, RB_CTRL));
}

/*
 * Setup prefetch unit registers. This is the interface between
 * hardware and driver list elements
 */
static void
myk_prefetch_init(
	struct gem_dev *dp, uint32_t qaddr, uint64_t addr, uint32_t last)
{
#if 0
	OUTL(dp, Y2_QADDR(qaddr, PREF_UNIT_CTRL_REG), PREF_UNIT_RST_SET);
#endif
	OUTL(dp, Y2_QADDR(qaddr, PREF_UNIT_CTRL_REG), PREF_UNIT_RST_CLR);
	OUTL(dp, Y2_QADDR(qaddr, PREF_UNIT_ADDR_LOW_REG), (uint32_t) addr);
	OUTL(dp, Y2_QADDR(qaddr, PREF_UNIT_ADDR_HI_REG), addr >> 32);
	OUTW(dp, Y2_QADDR(qaddr, PREF_UNIT_LAST_IDX_REG), last);
	OUTL(dp, Y2_QADDR(qaddr, PREF_UNIT_CTRL_REG), PREF_UNIT_OP_ON);

	FLSHL(dp, Y2_QADDR(qaddr, PREF_UNIT_CTRL_REG));
}

static void
myk_rx_set_checksum(struct gem_dev *dp, boolean_t hwcksum)
{
	uint_t			port = PORT(dp);
	struct msk_rx_desc	*rdp;
	struct myk_dev		*lp = (struct myk_dev *) dp->private;
	struct myk_port		*pp = &lp->descset[port];

	ASSERT(pp->rx_slot_head == 0);
	ASSERT(pp->rx_slot_tail == 0);

	rdp = &((struct msk_rx_desc *) RX_RING_BASE(dp))[
			SLOT(pp->rx_desc_tail, RX_RING_SIZE)];
	pp->rx_desc_tail++;

	rdp->msk_addr = (sizeof(struct ether_header) << 16) |
		    sizeof(struct ether_header);
	rdp->msk_control = OP_TCPSTART | HW_OWNER;

	OUTL(dp, Q_ADDR(RXQADDR(port), Q_CSR),
		hwcksum ? BMU_ENA_RX_CHKSUM : BMU_DIS_RX_CHKSUM);
}

static void
myk_status_desc_init(struct myk_dev *lp)
{
	uint_t		port = PORT(dp);
	struct gem_dev	*dp = lp->descset[port].dp;
	uint32_t	ramsize;
	uint32_t	rxspace;
	uint32_t	imask;
	int		cap;
	int		err;
	int		i;
	uint16_t	reg;
	int		thresh;

	DPRINTF(0, (CE_CONT, "!%s: %s: called", dp->name, __func__));

	bzero(STATUS_RING_BASE(dp), STATUS_LE_BYTES);
	lp->status_desc_tail = 0;

	OUTL(dp, STAT_CTRL, SC_STAT_RST_SET); /* b */
	OUTL(dp, STAT_CTRL, SC_STAT_RST_CLR); /* b */

	DPRINTF(0, (CE_CONT, "!%s: %s: status_le_base:0x%llx",
		dp->name, __func__, STATUS_RING_BASE_DMA(dp)));

	DPRINTF(0, (CE_CONT, "!%s: %s: rx_desc_base:0x%llx",
		dp->name, __func__, RX_RING_BASE_DMA(dp)));

	OUTL(dp, STAT_LIST_ADDR_LO, (uint32_t) STATUS_RING_BASE_DMA(dp));
	OUTL(dp, STAT_LIST_ADDR_HI, (uint32_t)(STATUS_RING_BASE_DMA(dp) >> 32));

	/* set the list last index */
	OUTW(dp, STAT_LAST_IDX, STATUS_RING_SIZE - 1);

	/* set Status-FIFO ISR watermark */ 
	if (lp->chip_id == CHIP_ID_YUKON_XL && lp->chip_rev == 0) {
		/* HWF_WA_DEV_43_418 */
		/* WA for dev. #4.3 */
		OUTW(dp, STAT_TX_IDX_TH, 10);
		/* WA for dev. #4.18 */
		OUTB(dp, STAT_FIFO_WM, 33);
		OUTB(dp, STAT_FIFO_ISR_WM, 7);
	} else {
		OUTW(dp, STAT_TX_IDX_TH, 10);
		OUTB(dp, STAT_FIFO_WM, 16);
		OUTB(dp, STAT_FIFO_ISR_WM, 16); /* HWF_WA_DEV_4109 */
		OUTL(dp, STAT_ISR_TIMER_INI, 0x0190);
	}
	OUTL(dp, STAT_TX_TIMER_INI, myk_us2clk(dp, 1000)); /* b */
	OUTL(dp, STAT_ISR_TIMER_INI, myk_us2clk(dp, 20));
	OUTL(dp, STAT_LEV_TIMER_INI, myk_us2clk(dp, 100));
#if 0
	myk_set_coalesce(dp);
#endif
	/* enable status unit */
	OUTL(dp, STAT_CTRL, SC_STAT_OP_ON); /* b */

	OUTB(dp, STAT_TX_TIMER_CTRL, TIM_START); /* b */
	OUTB(dp, STAT_LEV_TIMER_CTRL, TIM_START); /* b */
	OUTB(dp, STAT_ISR_TIMER_CTRL, TIM_START); /* b */
}

static int
myk_init_port(struct gem_dev *dp)
{
	uint_t		port = PORT(dp);
	struct myk_dev	*lp = (struct myk_dev *)dp->private;
	struct myk_port	*pp = &lp->descset[port];
	uint32_t	ramsize;
	uint32_t	rxspace;
	uint32_t	imask;
	int		cap;
	int		err;
	int		i;
	uint16_t	reg;
	int		thresh;

	DPRINTF(0, (CE_CONT, "!%s: %s: called", dp->name, __func__));
	ASSERT(mutex_owned(&dp->intrlock));

	/*
	 * Initialize GMAC first.
	 * Without this initialization, Rx MAC did not work as expected
	 * and Rx MAC garbled status LEs and it resulted in out-of-order
	 * or duplicated frame delivery which in turn showed very poor
	 * Rx performance.(I had to write a packet analysis code that
	 * could be embeded in driver to diagnose this issue.)
	 * I've spent almost 2 months to fix this issue. If I have had
	 * datasheet for Yukon II I wouldn't have encountered this. :-(
	 */
	OUTW(dp, GMAC_REG(port, GM_GP_CTRL),
	    GM_GPCR_SPEED_100 | GM_GPCR_SPEED_1000 | GM_GPCR_DUP_FULL);

	/* Dummy read the Interrupt Source Register. */
	FLSHB(dp, MR_ADDR(port, GMAC_IRQ_SRC));

#ifdef notyet
	/*
 	 * On dual port PCI-X card, there is an problem where status
	 * can be received out of order due to split transactions
	 */
	if (otherdev && netif_running(otherdev) &&
 	    (cap = pci_find_capability(hw->pdev, PCI_CAP_ID_PCIX))) {
 		struct sky2_port *osky2 = netdev_priv(otherdev);
 		uint16_t cmd;

 		cmd = sky2_pci_read16(dp, cap + PCI_X_CMD);
 		cmd &= ~PCI_X_CMD_MAX_SPLIT;
 		sky2_pci_write16(hw, cap + PCI_X_CMD, cmd);

 		sky2->rx_csum = 0;
 		osky2->rx_csum = 0;
 	}
#endif

	/* from sky2_mac_init(hw, port) */
#ifdef notyet
	if (lp->chip_id == CHIP_ID_YUKON_XL && lp->chip_rev == 0 && port == 1) {
		/* WA DEV_472 -- looks like crossed wires on port 2 */
		/* clear GMAC 1 Control reset */
		OUTB(dp, MR_ADDR(0, GMAC_CTRL), GMC_RST_CLR);
		do {
			OUTB(dp, MR_ADDR(1, GMAC_CTRL), GMC_RST_SET);
			OUTB(dp, MR_ADDR(1, GMAC_CTRL), GMC_RST_CLR);
		} while (gm_phy_read(dp, 1, PHY_MARV_ID0) != PHY_MARV_ID0_VAL ||
			 gm_phy_read(dp, 1, PHY_MARV_ID1) != PHY_MARV_ID1_Y2 ||
			 gm_phy_read(dp, 1, PHY_MARV_INT_MASK) != 0);
	}
#endif

	/* MIB clear */
	/* Set MIB Clear Counter Mode. */
	reg = INW(dp, GMAC_REG(port, GM_PHY_ADDR));
	OUTW(dp, GMAC_REG(port, GM_PHY_ADDR), reg | GM_PAR_MIB_CLR);
	/* Read all MIB Counters with Clear Mode set. */
	for (i = GM_MIB_CNT_BASE; i <= GM_TXE_FIFO_UR; i += 4) {
		(void) INW(dp, GMAC_REG(port, i));
	}
	/* Clear MIB Clear Counter Mode. */
	OUTW(dp, GMAC_REG(port, GM_PHY_ADDR), reg & ~GM_PAR_MIB_CLR);

	/* Disable FCS, promiscous mode */
	OUTW(dp, GMAC_REG(port, GM_RX_CTRL), GM_RXCR_CRC_DIS);

	/* Setup Transmit Control Register. */
	OUTW(dp, GMAC_REG(port, GM_TX_CTRL), TX_COL_THR(TX_COL_DEF));

	/* Setup Transmit Flow Control Register. */
	OUTW(dp, GMAC_REG(port, GM_TX_FLOW_CTRL), 0xffff);

	/* Setup Transmit Parameter Register. */
	OUTW(dp, GMAC_REG(port, GM_TX_PARAM),
		    TX_JAM_LEN_VAL(TX_JAM_LEN_DEF) |
		    TX_JAM_IPG_VAL(TX_JAM_IPG_DEF) |
		    TX_IPG_JAM_DATA(TX_IPG_JAM_DEF) |
		    TX_BACK_OFF_LIM(TX_BOF_LIM_DEF));

	/*
	 * Setup VLAN and MTU
	 */
	/* serial mode register */
	OUTW(dp, GMAC_REG(port, GM_SERIAL_MODE),
#ifdef GEM_CONFIG_VLAN
		GM_SMOD_JUMBO_ENA |
#endif
		DATA_BLIND_VAL(DATA_BLIND_DEF) |
		GM_SMOD_VLAN_ENA | IPG_DATA_VAL(IPG_DATA_DEF));

#ifdef notdef
	/* XXX - Set station address. later in set_rx_filter() */
#endif

	/* Disable interrupts for counter overflows */
	OUTW(dp, GMAC_REG(port, GM_TX_IRQ_MSK), 0);
	OUTW(dp, GMAC_REG(port, GM_RX_IRQ_MSK), 0);
	OUTW(dp, GMAC_REG(port, GM_TR_IRQ_MSK), 0);

	/* Configure Rx MAC FIFO. */
	OUTL(dp, MR_ADDR(port, RX_GMF_CTRL_T), GMF_RST_SET);
	OUTL(dp, MR_ADDR(port, RX_GMF_CTRL_T), GMF_RST_CLR);
	OUTL(dp, MR_ADDR(port, RX_GMF_CTRL_T), GMF_OPER_ON | GMF_RX_F_FL_ON);

	/* XXX - Set promiscuous mode. later in set_rx_filter() */
	/* XXX - set multicast filter. later in set_rx_filter() */

	/* Flush Rx MAC FIFO on any flow control or error */
	OUTL(dp, MR_ADDR(port, RX_GMF_FL_MSK), GMR_FS_ANY_ERR);

	/* Set threshold to 0xa (64 bytes) */
	OUTL(dp, MR_ADDR(port, RX_GMF_FL_THR), RX_GMF_FL_THR_DEF);

	/* Configure Tx MAC FIFO */
	OUTL(dp, MR_ADDR(port, TX_GMF_CTRL_T), GMF_RST_SET);
	OUTL(dp, MR_ADDR(port, TX_GMF_CTRL_T), GMF_RST_CLR);
	OUTL(dp, MR_ADDR(port, TX_GMF_CTRL_T), GMF_OPER_ON);

	/* set vlan */
	if ((dp->misc_flag & GEM_VLAN_HARD) != 0) {
		OUTL(dp, MR_ADDR(port, RX_GMF_CTRL_T), RX_VLAN_STRIP_ON);
		OUTL(dp, MR_ADDR(port, TX_GMF_CTRL_T), TX_VLAN_TAG_ON);
	} else {
		OUTL(dp, MR_ADDR(port, RX_GMF_CTRL_T), RX_VLAN_STRIP_OFF);
		OUTL(dp, MR_ADDR(port, TX_GMF_CTRL_T), TX_VLAN_TAG_OFF);
	}

	/* XXX - It seems STFW is requried for all cases. */
	OUTL(dp, MR_ADDR(port, TX_GMF_CTRL_T), TX_STFW_ENA);

	if (lp->chip_id == CHIP_ID_YUKON_EC_U) {
		OUTB(dp, MR_ADDR(port, RX_GMF_LP_THR), 768/8);
		OUTB(dp, MR_ADDR(port, RX_GMF_UP_THR), 1024/8);
#ifdef GEM_CONFIG_VLAN
		/*
		 * Can't sure the following code is needed as Yukon
		 * Yukon EC Ultra may not support jumbo frames.
		 *
		 * Set Tx GMAC FIFO Almost Empty Threshold.
		 */
		OUTL(dp, MR_ADDR(port, TX_GMF_AE_THR), 0x180);
		/* Disable Store & Forward mode for TX */
		OUTL(dp, MR_ADDR(port, TX_GMF_CTRL_T), TX_STFW_DIS);
#endif
	}

	/*
	 * Disable Force Sync bit and Alloc bit in Tx RAM interface
	 * arbiter as we don't use Sync Tx queue.
	 */
	OUTB(dp, MR_ADDR(port, TXA_CTRL),	
		TXA_DIS_FSYNC | TXA_DIS_ALLOC | TXA_STOP_RC);
	/* Enable the RAM Interface Arbiter. */
	OUTB(dp, MR_ADDR(port, TXA_CTRL), TXA_ENA_ARB);

	/*
	 * Determine available ram buffer space (in 4K blocks).
	 * Note: not sure about the FE setting below yet
	 */
	if (lp->chip_id == CHIP_ID_YUKON_FE) {
		ramsize = 4;
	} else {
		ramsize = INB(dp, B2_E_0);
	}
	/* Give transmitter one third (rounded up) */
	rxspace = ramsize - (ramsize + 2) / 3;
	myk_rambuffer_init(dp, RXQADDR(port), 0, rxspace);
	myk_rambuffer_init(dp, TXQADDR(port), rxspace, ramsize);

	DPRINTF(0, (CE_CONT, "%s: %s: ram buffer %dK (rx:%dK, tx:%dK)",
		dp->name, __func__,
		ramsize*4, rxspace*4, ramsize*4 - rxspace*4));

	/* Make sure SyncQ is disabled */
	OUTB(dp, RB_ADDR(port == 0 ? Q_XS1 : Q_XS2, RB_CTRL), RB_RST_SET);

	/*
	 * Setup Tx queue
	 */
	OUTL(dp, Q_ADDR(TXQADDR(port), Q_CSR), BMU_CLR_RESET);
	OUTL(dp, Q_ADDR(TXQADDR(port), Q_CSR), BMU_OPER_INIT);
	OUTL(dp, Q_ADDR(TXQADDR(port), Q_CSR), BMU_FIFO_OP_ON);
	OUTL(dp, Q_ADDR(TXQADDR(port), Q_WM),  0x600);

	/* Set almost empty threshold */
	/* XXX - need to check revision */
	if (lp->chip_id == CHIP_ID_YUKON_EC_U && lp->chip_rev == 0) {
		OUTW(dp, Q_ADDR(TXQADDR(port), Q_AL), 0x1a0);
	}

	/*
	 * Setup Rx queue
	 */
	OUTL(dp, Q_ADDR(RXQADDR(port), Q_CSR), BMU_CLR_RESET);
	OUTL(dp, Q_ADDR(RXQADDR(port), Q_CSR), BMU_OPER_INIT);
	OUTL(dp, Q_ADDR(RXQADDR(port), Q_CSR), BMU_FIFO_OP_ON);
	OUTL(dp, Q_ADDR(RXQADDR(port), Q_WM),  0x600);

	/* XXX - need to check revision */
	if (lp->chip_id == CHIP_ID_YUKON_EC_U && lp->chip_rev >= 2) {
		/* MAC Rx RAM Read is controlled by hardware */
		OUTL(dp, Q_ADDR(RXQADDR(port), Q_F), F_M_RX_RAM_DIS);
	}

	/*
	 * Setup Tx prefetch unit
	 */
	pp->tx_head = 0;
	pp->tx_tail = 0;
	myk_prefetch_init(dp, TXQADDR(port), dp->tx_ring_dma, TX_RING_SIZE-1);

	/*
	 * Setup Rx prefetch unit
	 */
	pp->rx_desc_head = 0;
	pp->rx_desc_tail = 0;
	pp->rx_slot_head = 0;
	pp->rx_slot_tail = 0;
	/* Disable Rx checksum offload and RSS hash. */
	OUTL(dp, Q_ADDR(RXQADDR(port), Q_CSR),
		BMU_DIS_RX_CHKSUM | BMU_DIS_RX_RSS_HASH);
	myk_prefetch_init(dp, RXQADDR(port),
			RX_RING_BASE_DMA(dp), RX_RING_SIZE - 1);

	/* put first command into rx descriptor ring */
	myk_rx_set_checksum(dp, (dp->misc_flag & GEM_CKSUM_PARTIAL) != 0);

	/*
	 * The receiver hangs if it receives frames larger than the
	 * packet buffer. As a workaround, truncate oversize frames, but
	 * the register is limited to 9 bits, so if you do frames > 2052
	 * you better get the MTU right!
	 */
	thresh = (dp->rx_buf_len - 8) / sizeof(uint32_t);
	if (thresh >= 512) {
		OUTL(dp, MR_ADDR(port, RX_GMF_CTRL_T), RX_TRUNC_OFF);
	} else {
		OUTW(dp, MR_ADDR(port, RX_GMF_TR_THR), thresh);
		OUTL(dp, MR_ADDR(port, RX_GMF_CTRL_T), RX_TRUNC_ON);
	}

	return (GEM_SUCCESS);
}

static int
myk_start_port(struct gem_dev *dp)
{
	uint_t		port = PORT(dp);
	struct myk_dev	*lp = (struct myk_dev *)dp->private;
	struct myk_port	*pp = &lp->descset[port];

	DPRINTF(0, (CE_CONT, "!%s: %s: called", dp->name, __func__));

	ASSERT(mutex_owned(&dp->intrlock));

	/* Enable Transmit FIFO Underrun */
	OUTB(dp, MR_ADDR(port, GMAC_IRQ_MSK), GMAC_DEF_MSK);

	OUTW(dp, GMAC_REG(port, GM_GP_CTRL),
		INW(dp, GMAC_REG(port, GM_GP_CTRL)) |
			GM_GPCR_RX_ENA | GM_GPCR_TX_ENA);

	/* Enable interrupts for the port */
	if ((dp->misc_flag & GEM_NOINTR) == 0) {
		/* Enable interrupts from phy/mac for port */
		lp->descset[port].hwe_imsk = 
			((port == 0) ? Y2_HWE_L1_MASK : Y2_HWE_L2_MASK) |
			Y2_IS_TIST_OV | Y2_IS_MST_ERR | Y2_IS_IRQ_STAT |
			Y2_IS_PCI_EXP;

		lp->descset[port].imsk = 
			PORTIRQ_MSK(port) | Y2_IS_HW_ERR | Y2_IS_STAT_BMU;

		OUTL(dp, B0_HWE_IMSK,
			lp->descset[0].hwe_imsk | lp->descset[1].hwe_imsk);
		FLSHL(dp, B0_HWE_IMSK);
		OUTL(dp, B0_IMSK,
			lp->descset[0].imsk | lp->descset[1].imsk);
		FLSHL(dp, B0_IMSK);
	}

	return (GEM_SUCCESS);
}

static int
myk_stop_port(struct gem_dev *dp)
{
	int		i;
	uint_t		port = PORT(dp);
	struct myk_dev	*lp = (struct myk_dev *) dp->private;
	struct myk_port	*pp = &lp->descset[port];
	uint_t		rxq = RXQADDR(port);

	DPRINTF(0, (CE_CONT, "!%s: %s: called", dp->name, __func__));

	/* Stop transmitter */
	OUTL(dp, Q_ADDR(TXQADDR(port), Q_CSR), BMU_STOP);
	FLSHL(dp, Q_ADDR(TXQADDR(port), Q_CSR));
	OUTL(dp, RB_ADDR(TXQADDR(port), RB_CTRL), RB_RST_SET | RB_DIS_OP_MD);

	OUTW(dp, GMAC_REG(port, GM_GP_CTRL),
		INW(dp, GMAC_REG(port, GM_GP_CTRL)) &
			~(GM_GPCR_TX_ENA | GM_GPCR_RX_ENA));

	OUTB(dp, MR_ADDR(port, GPHY_CTRL), GPC_RST_SET);
	OUTB(dp, MR_ADDR(port, GPHY_CTRL), GPC_RST_CLR); /* XXX */

	/* Workaround shared GMAC reset */
	if (lp->ports == 2) {
		if (!(lp->chip_id == CHIP_ID_YUKON_XL && lp->chip_rev == 0 &&
		     port == 0 && lp->descset[port^1].dp->mac_active)) {
			OUTB(dp, MR_ADDR(port, GMAC_CTRL), GMC_RST_SET);
			OUTB(dp, MR_ADDR(port, GMAC_CTRL), GMC_RST_CLR);
		}
	} else {
		OUTB(dp, MR_ADDR(port, GMAC_CTRL), GMC_RST_SET);
		OUTB(dp, MR_ADDR(port, GMAC_CTRL), GMC_RST_CLR);
	}

	/* Disable Force Sync bit and Enable Alloc bit */
	OUTB(dp, MR_ADDR(port, TXA_CTRL),
		    TXA_DIS_FSYNC | TXA_DIS_ALLOC | TXA_STOP_RC);

	/* Stop Interval Timer and Limit Counter of Tx Arbiter */
	OUTL(dp, MR_ADDR(port, TXA_ITI_INI), 0);
	OUTL(dp, MR_ADDR(port, TXA_LIM_INI), 0);

	/* Reset the PCI FIFO of the async Tx queue */
	OUTL(dp, Q_ADDR(TXQADDR(port), Q_CSR), BMU_RST_SET | BMU_FIFO_RST);

	/* Reset the Tx prefetch units */
	/* XXX - it implies  reset tx index */
	OUTL(dp, Y2_QADDR(TXQADDR(port),
			PREF_UNIT_CTRL_REG), PREF_UNIT_RST_SET);

	OUTL(dp, RB_ADDR(TXQADDR(port), RB_CTRL), RB_RST_SET);

	/*
	 * The RX Stop command will not work for Yukon-2 if the BMU does not
	 * reach the end of packet and since we can't make sure that we have
	 * incoming data, we must reset the BMU while it is not doing a DMA
	 * transfer. Since it is possible that the RX path is still active,
	 * the RX RAM buffer will be stopped first, so any possible incoming
	 * data will not trigger a DMA. After the RAM buffer is stopped, the
	 * BMU is polled until any DMA in progress is ended and only then it
	 * will be reset.
	 */

	/* disable the RAM Buffer receive queue */
	OUTB(dp, RB_ADDR(rxq, RB_CTRL), RB_DIS_OP_MD);

	i = 0;
	while (INB(dp, RB_ADDR(rxq, Q_RSL)) != INB(dp, RB_ADDR(rxq, Q_RL))) {
		if (i++ > 200) {
			cmn_err(CE_WARN,
				"%s: %s: timeout: receiver stop failed",
				dp->name, __func__);
			break;
		}
		drv_usecwait(10);
	}

	/*
	 * XXX - to reset rx index, we must issue prefetch unit reset while
	 * BMU_FIFO_RST is asserted
	 */
	OUTL(dp, Q_ADDR(rxq, Q_CSR), BMU_RST_SET | BMU_FIFO_RST);
	OUTL(dp, Y2_QADDR(RXQADDR(port),
			PREF_UNIT_CTRL_REG), PREF_UNIT_RST_SET); /* XXX */

	OUTB(dp, MR_ADDR(port, RX_GMF_CTRL_T), GMF_RST_SET);
	OUTB(dp, MR_ADDR(port, TX_GMF_CTRL_T), GMF_RST_SET);

	/* Disable port IRQ */
	lp->descset[port].hwe_imsk = 0;
	lp->descset[port].imsk = 0;
	OUTL(dp, B0_HWE_IMSK,
		lp->descset[0].hwe_imsk | lp->descset[1].hwe_imsk);
	FLSHL(dp, B0_HWE_IMSK);
	OUTL(dp, B0_IMSK,
		lp->descset[0].imsk | lp->descset[1].imsk);
	FLSHL(dp, B0_IMSK);

	return (GEM_SUCCESS);
}

static int
myk_attach_port(struct gem_dev *dp)	
{
	int		i;
	uint_t		port = PORT(dp);
	struct myk_dev	*lp = (struct myk_dev *)dp->private;
	struct myk_port	*pp = &lp->descset[port];

        /* read the mac address */
	for (i = 0; i < ETHERADDRL; i++) {
		dp->dev_addr.ether_addr_octet[i] =
			INB(dp, B2_MAC_1 + port * 8 + i);
	}

	dp->rx_buf_len =
		ROUNDUP2(dp->mtu + sizeof(struct ether_header) + 4, 8) + 8;

#ifdef GEM_CONFIG_CKSUM_OFFLOAD
	dp->misc_flag |= GEM_CKSUM_PARTIAL;
#endif
#ifdef GEM_CONFIG_VLAN
#ifdef CONFIG_HW_VLAN
	dp->misc_flag |= GEM_VLAN_HARD;
#else
	dp->misc_flag |= GEM_VLAN_SOFT;
#endif
#endif
	if (lp->ports > 1) {
		dp->misc_flag |= GEM_SOFTINTR;
	}
	return (GEM_SUCCESS);
}

static uint32_t
myk_mcast_hash(struct gem_dev *dp, uint8_t *addr)	/* OK */
{
	return (gem_ether_crc_be(addr, ETHERADDRL) & 0x3f);
}

static int
myk_set_rx_filter(struct gem_dev *dp)	
{
	int		i;
	uint16_t	reg;
	uint16_t	mode;
	uint32_t	hash[4];
	uint8_t		mac[ETHERADDRL];
	uint_t		port = PORT(dp);
	struct myk_dev	*lp = (struct myk_dev *) dp->private;
	struct myk_port	*pp = &lp->descset[port];

	DPRINTF(0, (CE_CONT, "!%s: %s: called", dp->name, __func__));

	hash[0] = hash[1] = hash[2] = hash[3] = 0;
	bcopy(dp->cur_addr.ether_addr_octet, mac, ETHERADDRL);

	if ((dp->rxmode & RXMODE_ENABLE) == 0) {
		mode = GM_RXCR_UCF_ENA | GM_RXCR_MCF_ENA;
		bzero(mac, ETHERADDRL);
	}
	else if ((dp->rxmode & RXMODE_PROMISC) != 0) {
		/* promiscuous */
		mode = 0;
	}
	else if ((dp->rxmode & RXMODE_ALLMULTI) != 0 || dp->mc_count > 32) {
		/* all multicast */
		mode = GM_RXCR_UCF_ENA;
	}
	else {
		mode = GM_RXCR_UCF_ENA | GM_RXCR_MCF_ENA;
		for (i = 0; i < dp->mc_count; i++) {
			uint32_t	h;

			h = dp->mc_list[i].hash;
			hash[h/16] |= 1 << (h & 16);
		}
	}
	/* we temporaly use promiscious mode */
	reg = INW(dp, GMAC_REG(port, GM_RX_CTRL)) &
			~(GM_RXCR_UCF_ENA | GM_RXCR_MCF_ENA);
	OUTW(dp, GMAC_REG(port, GM_RX_CTRL), reg);

	if ((mode & GM_RXCR_MCF_ENA) != 0) {
		/* update multicast filter */
		OUTW(dp, GMAC_REG(port, GM_MC_ADDR_H1), hash[0]);
		OUTW(dp, GMAC_REG(port, GM_MC_ADDR_H2), hash[1]);
		OUTW(dp, GMAC_REG(port, GM_MC_ADDR_H3), hash[2]);
		OUTW(dp, GMAC_REG(port, GM_MC_ADDR_H4), hash[3]);
	}

	/* update ethernet mac addresss */
	for (i = 0; i < ETHERADDRL; i++) {
		OUTB(dp, B2_MAC_1 + port * 8 + i, mac[i]);
		OUTB(dp, B2_MAC_2 + port * 8 + i, mac[i]);
	}

	/* virtual address for data */
	gma_set_addr(dp, port, GM_SRC_ADDR_2L,
			dp->cur_addr.ether_addr_octet);

	/* physical address: used for pause frames */
	gma_set_addr(dp, port, GM_SRC_ADDR_1L,
			dp->cur_addr.ether_addr_octet);

	/* update control register */
	OUTW(dp, GMAC_REG(port, GM_RX_CTRL), reg | mode);

	return (GEM_SUCCESS);
}

static void
myk_update_link_led(struct gem_dev *dp, int state)
{
}

static int
myk_set_media(struct gem_dev *dp)
{
	uint32_t	val;
	uint_t		port = PORT(dp);
	struct myk_dev	*lp = (struct myk_dev *) dp->private;
	struct myk_port	*pp = &lp->descset[port];

	DPRINTF(0, (CE_CONT, "!%s: %s: called", dp->name, __func__));

	val = INW(dp, GMAC_REG(port, GM_GP_CTRL));

	if (dp->mii_fixedmode) {
		/* select speed, flow control and duplex mode manually */
		val |= GM_GPCR_AU_ALL_DIS;
		switch (dp->speed) {
		case GEM_SPD_1000:
			val |= GM_GPCR_GIGS_ENA | GM_GPCR_SPEED_100;
			break;

		case GEM_SPD_100:
			val = (val & ~GM_GPCR_GIGS_ENA) | GM_GPCR_SPEED_100;
			break;

		case GEM_SPD_10:
			val &= ~(GM_GPCR_GIGS_ENA | GM_GPCR_SPEED_100);
			break;
		}

		/* select duplex mode */
		if (dp->full_duplex) {
			val |= GM_GPCR_DUP_FULL;
		} else {
			val &= ~GM_GPCR_DUP_FULL;
		}
	} else {
		/* select speed and duplex mode automatically */
		val &= ~GM_GPCR_AU_ALL_DIS;
	}

	/* select flow control */
	val |=  GM_GPCR_AU_FCT_DIS;
	switch (dp->flow_control) {
	case FLOW_CONTROL_NONE:
		/* disable Rx and Tx flow-control */
		val |= GM_GPCR_FC_TX_DIS | GM_GPCR_FC_RX_DIS;
		break;

	case FLOW_CONTROL_SYMMETRIC:
		/* Enable Rx and Tx flow-control */
		val &= ~(GM_GPCR_FC_TX_DIS | GM_GPCR_FC_RX_DIS);
		break;

	case FLOW_CONTROL_TX_PAUSE:
		/* disable Rx flow-control only */
		val = (val & ~GM_GPCR_FC_TX_DIS) | GM_GPCR_FC_RX_DIS;
		break;

	case FLOW_CONTROL_RX_PAUSE:
		/* disable Tx flow-control only */
		val = (val & ~GM_GPCR_FC_RX_DIS) | GM_GPCR_FC_TX_DIS;
		break;
	}

	if (dp->flow_control != FLOW_CONTROL_NONE) {
		OUTB(dp, MR_ADDR(port, GMAC_CTRL), GMC_PAUSE_ON);
	} else {
		OUTB(dp, MR_ADDR(port, GMAC_CTRL), GMC_PAUSE_OFF);
	}
#ifdef NEVER
	val |= GM_GPCR_RX_ENA | GM_GPCR_TX_ENA;
#endif
	OUTW(dp, GMAC_REG(port, GM_GP_CTRL), val);
	FLSHW(dp, GMAC_REG(port, GM_GP_CTRL));

	return (GEM_SUCCESS);
}

static int
myk_get_stats(struct gem_dev *dp)
{
#ifdef notdef
	struct sky2_port *sky2 = netdev_priv(dev);
	u64 data[13];

	sky2_phy_stats(sky2, data, ARRAY_SIZE(data));

	sky2->net_stats.tx_bytes = data[0];
	sky2->net_stats.rx_bytes = data[1];
	sky2->net_stats.tx_packets = data[2] + data[4] + data[6];
	sky2->net_stats.rx_packets = data[3] + data[5] + data[7];
	sky2->net_stats.multicast = data[3] + data[5];
	sky2->net_stats.collisions = data[10];
	sky2->net_stats.tx_aborted_errors = data[12];
#endif
	return (GEM_SUCCESS);
}

/*
 * discriptor  manupiration
 */
static int
myk_tx_desc_write(struct gem_dev *dp, int slot,
		ddi_dma_cookie_t *dmacookie, int frags, uint32_t flag)
{
	int		i;
	uint64_t	vtag;
	int		used;
	uint32_t	opcode;
	uint32_t	hiaddr;
	uint32_t	hiaddr2;
	uint32_t	ctrl;
	uint_t		port = PORT(dp);
	struct myk_dev	*lp = (struct myk_dev *)dp->private;
	struct myk_port	*pp = &lp->descset[port];
	ddi_dma_cookie_t	*dcp;
	struct msk_tx_desc	*tdp;

#if DEBUG_LEVEL > 2
	/* force to cause interrupt upon tx completion */
	flag |= GEM_TXFLAG_INTR;
#endif

#if DEBUG_LEVEL > 2
	cmn_err(CE_CONT,
		"%s: %s: seqnum: %d, slot %d, frags: %d flag: %d",
		dp->name, __func__, dp->tx_desc_tail, slot, frags, flag);

	for (i = 0; i < frags; i++) {
		cmn_err(CE_CONT, "%d: addr: 0x%x, len: 0x%x",
			i, dmacookie[i].dmac_address, dmacookie[i].dmac_size);
	}
#endif
	used = 0;
	dcp  = dmacookie;

	/* Send high bits if changed or crosses boundary */
	ctrl = HW_OWNER;
	hiaddr = (uint32_t) (dcp->dmac_laddress >> 32);
	hiaddr2 = (dcp->dmac_laddress + dcp->dmac_size) >> 32;
	vtag = (flag & GEM_TXFLAG_VTAG);

	if (hiaddr != pp->tx_hiaddr || hiaddr != hiaddr2 || vtag != 0) {

		tdp = &((struct msk_tx_desc *) dp->tx_ring)[
				SLOT(slot + (used++), TX_RING_SIZE)];

		opcode = OP_ADDR64 | HW_OWNER;
#ifdef CONFIG_HW_VLAN
		if ((vtag >>= GEM_TXFLAG_VTAG_SHIFT) != 0) {
			ctrl |= INS_VLAN;
			opcode |= OP_VLAN | vtag;
		}
#endif
		tdp->msk_addr    = LE32(hiaddr);
		tdp->msk_control = LE32(opcode);

		pp->tx_hiaddr = hiaddr2;
	}

	/* Handle TCP checksum offload */
	if ((flag & (GEM_TXFLAG_UDP | GEM_TXFLAG_TCP)) != 0) {
		uint32_t	start;

		start = sizeof(struct ether_header) + sizeof(struct ip);

		/* XXX - no need to increase offset for HW vlan */
		if ((flag & GEM_TXFLAG_SWVTAG) != 0) {
			start += 4;
		}

		tdp = &((struct msk_tx_desc *) dp->tx_ring)[
				SLOT(slot + (used++), TX_RING_SIZE)];

		ctrl |= CALSUM | WR_SUM | INIT_SUM | LOCK_SUM;	/*was =, bug? */
		if ((flag & GEM_TXFLAG_UDP) != 0) {
			ctrl |= UDPTCP;
			start = (start << 16) |
				(start + offsetof(struct udphdr, uh_sum));
		} else {
			start = (start << 16) |
				(start + offsetof(struct tcphdr, th_sum));
		}
		/* number of packets: 1, use initial checksum in the packet area */
		tdp->msk_control = LE32(OP_TCPLISW | HW_OWNER | (1 << 16) | 0);
		tdp->msk_addr    = LE32(start);
	}

	tdp = &((struct msk_tx_desc *) dp->tx_ring)[
			SLOT(slot + (used++), TX_RING_SIZE)];

	tdp->msk_addr    = LE32(dcp->dmac_address);
	tdp->msk_control = LE32(OP_PACKET | ctrl | dcp->dmac_size);
	dcp++;

	for (i = 1; i < frags; i++) {
		hiaddr = (uint32_t) (dcp->dmac_laddress >> 32);
		hiaddr2 = (dcp->dmac_laddress + dcp->dmac_size) >> 32;

		if (hiaddr != pp->tx_hiaddr || hiaddr != hiaddr2) {
			tdp = &((struct msk_tx_desc *) dp->tx_ring)[
				SLOT(slot + (used++), TX_RING_SIZE)];

			pp->tx_hiaddr = hiaddr2;

			tdp->msk_addr    = LE32(hiaddr);
			tdp->msk_control = LE32(OP_ADDR64 | HW_OWNER);
		}

		tdp = &((struct msk_tx_desc *) dp->tx_ring)[
				SLOT(slot + (used++), TX_RING_SIZE)];

		tdp->msk_addr    = LE32(dcp->dmac_address);
		tdp->msk_control = LE32(OP_BUFFER | ctrl | dcp->dmac_size);
		dcp++;
	}

	/* close the fragment list */
	ctrl = EOP;
	if ((flag & GEM_TXFLAG_INTR) != 0) {
 		ctrl |= FRC_STAT;
	}
	tdp->msk_control |= LE32(ctrl);

	return (used);
}

static void
myk_tx_start(struct gem_dev *dp, int start_slot, int nslot)
{
	uint_t		port = PORT(dp);
	struct myk_dev	*lp = (struct myk_dev *) dp->private;
	struct myk_port	*pp = &lp->descset[port];

	/* set new tail */
	pp->tx_tail = SLOT(start_slot + nslot, TX_RING_SIZE);

	gem_tx_desc_dma_sync(dp, start_slot, nslot, DDI_DMA_SYNC_FORDEV);

	OUTL(dp, Y2_QADDR(TXQADDR(port), PREF_UNIT_PUT_IDX_REG), pp->tx_tail);
}

static void
myk_rx_desc_write(struct gem_dev *dp, int slot,
	    ddi_dma_cookie_t *dmacookie, int frags)
{
	int		i;
	uint32_t	hiaddr;
	uint32_t	hiaddr2;
	uint_t		port = PORT(dp);
	struct myk_dev	*lp = (struct myk_dev *) dp->private;
	struct myk_port	*pp = &lp->descset[port];
	struct msk_rx_desc	*rdp;

	/*
	 * we always consume two descriptors
	 */
#if DEBUG_LEVEL > 1
	cmn_err(CE_CONT, "%s: %s seqnum: %d, slot %d, frags: %d",
		dp->name, __func__, dp->rx_active_tail, slot, frags);
	for (i = 0; i < frags; i++) {
		cmn_err(CE_CONT, "  frag: %d addr: 0x%x, len: 0x%x",
			i, dmacookie[i].dmac_address, dmacookie[i].dmac_size);
	}
#endif
	hiaddr = dmacookie->dmac_laddress >> 32;
	hiaddr2 = (dmacookie->dmac_laddress + dmacookie->dmac_size) >> 32;

	if (pp->rx_hiaddr != hiaddr || hiaddr2 != hiaddr) {
		rdp = &((struct msk_rx_desc *) RX_RING_BASE(dp))[
				SLOT((pp->rx_desc_tail++), RX_RING_SIZE)];
		rdp->msk_addr    = LE32(dmacookie->dmac_laddress >> 32);
		rdp->msk_control = LE32(OP_ADDR64 | HW_OWNER);

		pp->rx_hiaddr = hiaddr2;
	}

	rdp = &((struct msk_rx_desc *) RX_RING_BASE(dp))[
			SLOT((pp->rx_desc_tail++), RX_RING_SIZE)];
	rdp->msk_addr = LE32(dmacookie->dmac_address);
	rdp->msk_control = LE32(OP_PACKET | HW_OWNER | dmacookie->dmac_size);
}

static void
myk_rx_start(struct gem_dev *dp, int start_slot, int nslot)
{
	uint_t		port = PORT(dp);
	struct myk_dev	*lp = (struct myk_dev *) dp->private;
	struct myk_port	*pp = &lp->descset[port];

	DPRINTF(1, (CE_CONT, "%s: %s: start:%d, nslot:%d, rx_desc(%d, %d)",
		dp->name, __func__, start_slot, nslot,
		pp->rx_desc_head, pp->rx_desc_tail));

	myk_rx_desc_dma_sync(dp,
		SLOT(pp->rx_desc_head, RX_RING_SIZE),
		pp->rx_desc_tail - pp->rx_desc_head,
		DDI_DMA_SYNC_FORDEV);

	pp->rx_desc_head = pp->rx_desc_tail,
	
	OUTW(dp, Y2_QADDR(RXQADDR(port), PREF_UNIT_PUT_IDX_REG),
			SLOT(pp->rx_desc_tail, RX_RING_SIZE));
}

static uint_t
myk_tx_desc_stat(struct gem_dev *dp, int slot, int ndesc)
{
	uint_t		port = PORT(dp);
	struct myk_dev	*lp = (struct myk_dev *) dp->private;
	struct myk_port	*pp = &lp->descset[port];

	slot = SLOT(slot + ndesc - 1, TX_RING_SIZE);

	if (INSIDE(slot, pp->tx_head, pp->tx_tail)) {
		return (0);
	}
	DPRINTF(0, (CE_CONT,
		"%s: %s: slot:%d, [%d, %d]",
		dp->name, __func__, slot, pp->tx_head, pp->tx_tail));

	return (GEM_TX_DONE);
}

static uint64_t
myk_rx_desc_stat(struct gem_dev *dp, int slot, int ndesc)
{
	uint64_t	flag = GEM_RX_DONE;
	uint_t		port = PORT(dp);
	struct myk_dev	*lp = (struct myk_dev *) dp->private;
	struct myk_port	*pp = &lp->descset[port];
	struct rx_soft_desc	*rdp;

	ASSERT(ndesc == 1);

	if (slot == SLOT(pp->rx_slot_tail, RX_RING_SIZE)) {
		/* not received yet */
		return (0);
	}

	/* remove the head */
	ASSERT(slot == SLOT(pp->rx_slot_head, RX_RING_SIZE));
	pp->rx_slot_head++;

	rdp = &pp->rx_slot[slot];

	DPRINTF(1, (CE_CONT,
		"%s: %s: slot:%d, len:0x%x, stat:0x%x",
		dp->name, __func__, slot, rdp->rx_len, rdp->rx_stat));

	if ((rdp->rx_stat & (GMR_FS_ANY_ERR | GMR_FS_RX_OK)) != GMR_FS_RX_OK) {
		DPRINTF(0, (CE_NOTE,
			"%s: %s: rx error, status 0x%x length %d",
			dp->name, __func__, rdp->rx_stat, rdp->rx_len));

		/* error packet */
		dp->stats.errrcv++;

		if ((rdp->rx_stat & (GMR_FS_LONG_ERR | GMR_FS_UN_SIZE)) != 0) {
			dp->stats.frame_too_long++;
			/* dp->stats.runt++; */
		}
		if ((rdp->rx_stat & GMR_FS_FRAGMENT) != 0) {
			dp->stats.frame++;
		}
		if ((rdp->rx_stat & GMR_FS_CRC_ERR) != 0) {
			dp->stats.crc++;
		}
		if ((rdp->rx_stat & GMR_FS_RX_FF_OV) != 0) {
			dp->stats.overflow++;
		}
		flag = GEM_RX_ERR;
	} else {
		flag |= ((uint64_t) rdp->rx_cksum) << GEM_RX_CKSUM_SHIFT;
		flag |= ((uint64_t) rdp->rx_tag) << GEM_RX_VTAG_SHIFT;
	}

	rdp->rx_cksum = 0;
	rdp->rx_tag   = 0;
	rdp->rx_stat  = 0;

	return (flag | rdp->rx_len);
}

static void
myk_tx_desc_init(struct gem_dev *dp, int slot)
{
	bzero(&((struct msk_tx_desc *) dp->tx_ring)[slot],
		sizeof(struct msk_tx_desc));
}

static void
myk_rx_desc_init(struct gem_dev *dp, int slot)
{
	bzero(&((struct msk_rx_desc *) RX_RING_BASE(dp))[slot],
		sizeof(struct msk_rx_desc));
}

/*
 * Device depend interrupt handler
 */
static void
myk_hw_error(struct gem_dev *dp, uint_t port, uint32_t status)
{
	DPRINTF(0, (CE_NOTE, "%s: hw error interrupt status 0x%x",
		       dp->name, status));

	if ((status & Y2_IS_PAR_RD1) != 0) {
		cmn_err(CE_WARN, "%s: ram data read parity error", dp->name);
		/* Clear IRQ */
		OUTW(dp, RAM_BUFFER(port, B3_RI_CTRL), RI_CLR_RD_PERR);
	}

	if ((status & Y2_IS_PAR_WR1) != 0) {
		cmn_err(CE_WARN, "%s: ram data write parity error", dp->name);
		OUTW(dp, RAM_BUFFER(port, B3_RI_CTRL), RI_CLR_WR_PERR);
	}

	if ((status & Y2_IS_PAR_MAC1) != 0) {
		cmn_err(CE_WARN, "%s: MAC parity error", dp->name);
		OUTB(dp, MR_ADDR(port, TX_GMF_CTRL_T), GMF_CLI_TX_PE);
	}

	if ((status & Y2_IS_PAR_RX1) != 0) {
		cmn_err(CE_WARN, "%s: RX parity error", dp->name);
		OUTL(dp, Q_ADDR(RXQADDR(port), Q_CSR), BMU_CLR_IRQ_PAR);
	}

	if ((status & Y2_IS_TCP_TXA1) != 0) {
		cmn_err(CE_WARN, "%s: TCP segmentation error", dp->name);
		OUTL(dp, Q_ADDR(TXQADDR(port), Q_CSR), BMU_CLR_IRQ_TCP);
	}
}

static void
myk_hw_intr(struct gem_dev *dp)
{
	uint32_t	status;

	status = INL(dp, B0_HWE_ISRC);

	if ((status & Y2_IS_TIST_OV) != 0) {
		OUTB(dp, GMAC_TI_ST_CTRL, GMT_ST_CLR_IRQ);
	}

	if (status & (Y2_IS_MST_ERR | Y2_IS_IRQ_STAT)) {
		uint16_t	pci_err;

		pci_err = myk_pci_read16(dp, PCI_CONF_STAT);
		cmn_err(CE_WARN, "%s: pci hw error (0x%x)", dp->name, pci_err);

		OUTB(dp, B2_TST_CTRL1, TST_CFG_WRITE_ON);

		myk_pci_write16(dp, PCI_CONF_STAT, pci_err |
		    PCI_STAT_PERROR | PCI_STAT_S_SYSERR | PCI_STAT_R_MAST_AB |
		    PCI_STAT_R_TARG_AB | PCI_STAT_S_PERROR); /*b*/

		OUTB(dp, B2_TST_CTRL1, TST_CFG_WRITE_OFF);
	}

	if ((status & Y2_IS_PCI_EXP) != 0) {
		/* PCI-Express uncorrectable Error occurred */
		uint32_t	pex_err;

		pex_err = myk_pci_read32(dp, PEX_UNC_ERR_STAT);

		cmn_err(CE_WARN, "%s: pci express error (0x%x)",
			dp->name, pex_err);

		/* clear the interrupt */
		OUTL(dp, B2_TST_CTRL1, TST_CFG_WRITE_ON);
		myk_pci_write32(dp, PEX_UNC_ERR_STAT, 0xffffffff);
		OUTL(dp, B2_TST_CTRL1, TST_CFG_WRITE_OFF);

		if ((pex_err & PEX_FATAL_ERRORS) != 0) {
			OUTL(dp, B0_HWE_IMSK,
				INL(dp, B0_HWE_IMSK) & ~Y2_IS_PCI_EXP);
		}
	}

	if ((status & Y2_HWE_L1_MASK) != 0) {
		myk_hw_error(dp, 0, status);
	}

	status >>= 8;
	if ((status & Y2_HWE_L1_MASK) != 0) {
		myk_hw_error(dp, 1, status);
	}
}

/* Interrupt from PHY */
static void
myk_phy_intr(struct gem_dev *dp, uint_t port)
{
	uint16_t	istatus;
	uint16_t	phystat;

	/* XXX - Do we need mutex ? */
	istatus = myk_mii_read(dp, PHY_MARV_INT_STAT);
	phystat = myk_mii_read(dp, PHY_MARV_PHY_STAT);

	DPRINTF(0, (CE_CONT, "%s: port%d: phy interrupt status 0x%x 0x%x",
		       dp->name, port, istatus, phystat));
#ifdef DEBUG_LEVEL
	if ((istatus & PHY_M_IS_AN_COMPL) != 0) {
		DPRINTF(0, (CE_NOTE, "%s: %s: autonegotiation completed",
			dp->name, __func__));
		goto out;
	}

	if ((istatus & PHY_M_IS_LSP_CHANGE) != 0) {
		DPRINTF(0, (CE_NOTE, "%s: %s: link speed changed",
			dp->name, __func__));
	}

	if ((istatus & PHY_M_IS_DUP_CHANGE) != 0) {
		DPRINTF(0, (CE_NOTE, "%s: %s: duplex changed",
			dp->name, __func__));
	}

	if ((istatus & PHY_M_IS_LST_CHANGE) != 0) {
		DPRINTF(0, (CE_NOTE, "%s: %s: link status changed",
			dp->name, __func__));
	}
out:
	return;
#endif
}

static void
myk_mac_intr(struct gem_dev *dp, unsigned port)
{
	uint8_t		status;
	struct myk_dev	*lp = (struct myk_dev *) dp->private;

	status = INB(dp, MR_ADDR(port, GMAC_IRQ_SRC));

	DPRINTF(0, (CE_CONT,  "%s: %s: mac interrupt status 0x%x",
		       dp->name, __func__, status));

	if ((status & GM_IS_RX_FF_OR) != 0) {
		/* ++sky2->net_stats.rx_fifo_errors; */
		OUTB(dp, MR_ADDR(port, RX_GMF_CTRL_T), GMF_CLI_RX_FO);
	}

	if ((status & GM_IS_TX_FF_UR) != 0) {
		/* ++sky2->net_stats.tx_fifo_errors; */
		OUTB(dp, MR_ADDR(port, TX_GMF_CTRL_T), GMF_CLI_TX_FU);
	}
}

/* This should never happen it is a fatal situation */
static void
myk_descriptor_error(struct gem_dev *dp, uint_t port,
				  const char *rxtx, uint32_t mask)
{
	uint32_t	imask;
	struct myk_dev	*lp = (struct myk_dev *) dp->private;

	cmn_err(CE_WARN, "%s: %s: %s: descriptor error (hardware problem)",
		dp->name, __func__ , rxtx);

	OUTL(dp, B0_IMSK, INL(dp, B0_IMSK) & ~mask);
}

/*
 * Real hardware interrupt service routine
 */
#define	V_ISR_HWERR	0x0008
#define	V_ISR_LSC	0x0004
#define	V_ISR_TX	0x0002
#define	V_ISR_RX	0x0001
static u_int
myk_interrupt(struct gem_dev *dp0)
{
	int		i;
	u_int		hwidx;
	uint32_t	status;
	u_int		isr[2];
	int		slot;
	struct myk_dev	*lp = (struct myk_dev *) dp0->private;
	struct myk_port	*pp;

	/* Reading this mask interrupts as side effect */
	status = INL(dp0, B0_Y2_SP_ISRC2);

	DPRINTF(2, (CE_CONT,"!%s: time:%d %s: status:%x",
		dp0->name, ddi_get_lbolt(), __func__, status));

	if (status == 0 || status == 0xffffffff) {
		/* not for us */
#if 0
		OUTL(dp0, B0_Y2_SP_ICR, 2);
#endif
		return (DDI_INTR_UNCLAIMED);
	}

	DPRINTF(1, (CE_CONT,"!%s: time:%d %s: isrc2:%b, isrc:%x, isrc3:%x",
		dp0->name, ddi_get_lbolt(), __func__, status, IS_BITS,
		INL(dp0, B0_ISRC), INL(dp0, B0_Y2_SP_ISRC3)));

	if ((status & (lp->descset[0].imsk | lp->descset[1].imsk)) == 0) {
		/* not for us */
		return (DDI_INTR_UNCLAIMED);
	}

	status &= lp->descset[0].imsk | lp->descset[1].imsk;

	/* ack to the interrupts */
	OUTL(dp0, B0_Y2_SP_ICR, 2);

	/* clear virtual interrupt source resgister */
	isr[0] = isr[1] = 0;

	if ((status & Y2_IS_HW_ERR) != 0) {
		myk_hw_intr(dp0);
		isr[0] |= V_ISR_HWERR;
		isr[1] |= V_ISR_HWERR;
	}
	if ((status & Y2_IS_IRQ_PHY1) != 0) {
		pp = &lp->descset[0];
		myk_phy_intr(pp->dp, 0);
		isr[0] |= V_ISR_LSC;
	}
	if ((status & Y2_IS_IRQ_PHY2) != 0) {
		pp = &lp->descset[1];
		myk_phy_intr(pp->dp, 1);
		isr[1] |= V_ISR_LSC;
	}
	if ((status & Y2_IS_IRQ_MAC1) != 0) {
		pp = &lp->descset[0];
		myk_mac_intr(pp->dp, 0);
		isr[0] |= V_ISR_HWERR;
	}
	if ((status & Y2_IS_IRQ_MAC2) != 0) {
		pp = &lp->descset[1];
		myk_mac_intr(pp->dp, 1);
		isr[1] |= V_ISR_HWERR;
	}
	if ((status & Y2_IS_CHK_RX1) != 0) {
		pp = &lp->descset[0];
		myk_descriptor_error(pp->dp, 0, "receive", Y2_IS_CHK_RX1);
		isr[0] |= V_ISR_HWERR;
		cmn_err(CE_CONT,
			"%s: rx_desc (%d, %d)",
			pp->dp->name, pp->rx_desc_head,
			pp->rx_desc_tail);
	}
	if ((status & Y2_IS_CHK_RX2) != 0) {
		pp = &lp->descset[1];
		myk_descriptor_error(pp->dp, 1, "receive", Y2_IS_CHK_RX2);
		isr[1] |= V_ISR_HWERR;
	}
	if ((status & Y2_IS_CHK_TXA1) != 0) {
		pp = &lp->descset[0];
		myk_descriptor_error(pp->dp, 0, "transmit", Y2_IS_CHK_TXA1);
		isr[0] |= V_ISR_HWERR;
	}
	if ((status & Y2_IS_CHK_TXA2) != 0) {
		pp = &lp->descset[1];
		myk_descriptor_error(pp->dp, 1, "transmit", Y2_IS_CHK_TXA2);
		isr[1] |= V_ISR_HWERR;
	}

	/*
	 * Interpret status descriptors
	 */
	slot = SLOT(lp->status_desc_tail, STATUS_RING_SIZE);

	OUTL(dp0, STAT_CTRL, SC_STAT_CLR_IRQ);

	hwidx = INW(dp0, STAT_PUT_IDX);

	DPRINTF(1, (CE_CONT, "%s: %s: hwidx:%d, desc_tail:%d",
		dp0->name, __func__, hwidx, lp->status_desc_tail));

	if ((i = (hwidx - slot)) < 0) {
		i += STATUS_RING_SIZE;
	}
	myk_status_desc_dma_sync(dp0, slot, i, DDI_DMA_SYNC_FORDEV);

	while (slot != hwidx) {
		u_int	control;
		u_int	port;
		u_int	length;
		u_int	pos;
		u_int	op;
		struct msk_stat_desc	*sdp;

		/* get pointer to the current status descriptor */
		sdp = &((struct msk_stat_desc *) STATUS_RING_BASE(dp0))[slot];

#define	SDES_OP		0x7f000000
#define		SDES_OP_SHIFT	24
#define	SDES_LINK	0x00ff0000
#define		SDES_LINK_SHIFT	16
#define	SDES_LENGTH	0x0000ffff

		control = LE32(sdp->msk_control);
		port = (control & SDES_LINK) >> SDES_LINK_SHIFT;
		op = (control & SDES_OP) >> SDES_OP_SHIFT;
		length = control & SDES_LENGTH;
		status = LE32(sdp->msk_status);

		if (port < 0 || port >= lp->ports) {
			cmn_err(CE_WARN,
			"%s: %s: status entry hw error:"
			"port %d shoube be 0 or 1, slot:%d, slot_tail:%d",
			dp0->name, __func__, port, slot, lp->status_desc_tail);
			goto x;
		}

		/* advance status tail */
		slot = SLOT(slot + 1, STATUS_RING_SIZE);
		lp->status_desc_tail++;
		ASSERT(slot == SLOT(lp->status_desc_tail, STATUS_RING_SIZE));

		/* interpret the status descriptor */

		DPRINTF(1, (CE_CONT,
			"%s: %s: idx:%d op:0x%x len:0x%x, stat:0x%x",
			dp0->name, __func__, hwidx, op, length, status));

		switch (op) {
		case OP_RXSTAT >> SDES_OP_SHIFT:
			DPRINTF(1, (CE_CONT,
			"%s: %s: idx:%d OP_RXSTAT(%x) len:0x%x, stat:0x%x",
				dp0->name, __func__, hwidx, op,
				length, status));

			pp = &lp->descset[port];
			i = SLOT(pp->rx_slot_tail, RX_RING_SIZE);
			pp->rx_slot[i].rx_stat = status;
			pp->rx_slot[i].rx_len  = length;
			pp->rx_slot_tail++;
			isr[port] |= V_ISR_RX;
			break;
#ifdef CONFIG_HW_VLAN
		case OP_RXVLAN >> SDES_OP_SHIFT:
		case OP_RXCHKSVLAN >> SDES_OP_SHIFT:
			DPRINTF(1, (CE_CONT,
			"%s: %s: idx:%d %s(%x)  tag:0x%x, cksum:0x%x",
				dp0->name, __func__, hwidx,
				(op == OP_RXVLAN) ?
					"OP_RXVLAN" : "OP_RXCHKSVLAN",
				op, length, status));

			pp = &lp->descset[port];
			i = SLOT(pp->rx_slot_tail, RX_RING_SIZE);
			pp->rx_slot[i].rx_tag = length;
			if (op == OP_RXCHKSVLAN) {
				pp->rx_slot[i].rx_cksum = status;
			}
			break;
#endif
		case OP_RXCHKS >> SDES_OP_SHIFT:
			DPRINTF(1, (CE_CONT,
			"%s: %s: idx:%d OP_RXCHKS(%x)  control:0x%x stat:0x%x",
			    dp0->name, __func__, hwidx, op, control, status));

			pp = &lp->descset[port];
			i = SLOT(pp->rx_slot_tail, RX_RING_SIZE);
			pp->rx_slot[i].rx_cksum = status;
			break;

		case OP_TXINDEXLE >> SDES_OP_SHIFT:
			/* TX index reports status for both ports */
			ASSERT(TX_RING_SIZE <= 0x1000);

			pp = &lp->descset[0];
			pos = status & 0xfff;
			if (pp->tx_head != pos) {
				pp->tx_head = pos;
				isr[0] |= V_ISR_TX;
			}

			if (lp->ports == 2) {
				/* for additional second port */
				pp = &lp->descset[1];
				pos = ((length << 8) | (status >> 24)) & 0xfff;
				if (pp->tx_head != pos) {
					pp->tx_head = pos;
					isr[1] |= V_ISR_TX;
				}
			}
			DPRINTF(1, (CE_CONT,
			"%s: %s: idx:%d OP_TXINDEXLE(%x)  p0:%d p1:%d",
				dp0->name, __func__, hwidx, op,
				status & 0xfff,
				((length << 8) | (status >> 24)) & 0xfff));

			break;

		default:
			cmn_err(CE_WARN,
				"%s: %s: idx:%d unknown status opcode 0x%x",
				dp0->name, __func__, hwidx, op);
			goto x;
		}
	}
x:
	/* kick software interrupt handler for each port */
	for (i = 0; i < lp->ports; i++) {
		lp->descset[i].isr = isr[i];
		if (isr[i] != 0 && lp->descset[i].dp->soft_id) {
			ddi_trigger_softintr(lp->descset[i].dp->soft_id);
		}
	}

	return (DDI_INTR_CLAIMED);
}

/*
 * soft interrupt service routine
 */
static u_int
myk_soft_intr(struct gem_dev *dp)
{
	uint32_t	flags = 0;
	u_int		isr;
	boolean_t	need_to_reset = B_FALSE;
	struct myk_dev	*lp = (struct myk_dev *) dp->private;
	struct myk_port	*pp = &lp->descset[PORT(dp)];

	DPRINTF(2, (CE_CONT,"!%s: time:%d %s: isr:%x",
		dp->name, ddi_get_lbolt(), __func__, pp->isr));

	ASSERT(mutex_owned(&dp->intrlock));

	if ((dp->misc_flag & GEM_SOFTINTR) == 0) {
		if (myk_interrupt(dp) != DDI_INTR_CLAIMED) {
			return (DDI_INTR_UNCLAIMED);
		}
	}

	if ((isr = pp->isr) == 0) {
		return (DDI_INTR_UNCLAIMED);
	}

	pp->isr = 0;

	if (!dp->mac_active) {
		return (DDI_INTR_UNCLAIMED);
	}

	if ((isr & V_ISR_RX) != 0) {
		(void) gem_receive(dp);
	}

	if ((isr & V_ISR_TX) != 0) {
		if (gem_tx_done(dp)) {
			flags = INTR_RESTART_TX;
		}
	}

	if ((isr & V_ISR_LSC) != 0) {
		if (gem_mii_link_check(dp)) {
			flags = INTR_RESTART_TX;
		}
	}

	if ((isr & V_ISR_HWERR) != 0) {
		need_to_reset = B_TRUE;
	}

	if (need_to_reset) {
		gem_restart_nic(dp, GEM_RESTART_KEEP_BUF);
		flags = INTR_RESTART_TX;
	}

	return (DDI_INTR_CLAIMED | flags);
}

/*
 * HW depend MII routine
 */
static void
myk_mii_sync(struct gem_dev *dp)
{
	/* do nothing */
}

static uint16_t
myk_mii_read(struct gem_dev *dp, uint_t reg)
{
	int		i;
	uint16_t	val;
	uint_t		port = PORT(dp);
	struct myk_dev	*lp = (struct myk_dev *) dp->private;

	ASSERT(port == 0);

	OUTW(dp, GMAC_REG(port, GM_SMI_CTRL),
		GM_SMI_CT_PHY_AD(dp->mii_phy_addr) |
		GM_SMI_CT_REG_AD(reg) | GM_SMI_CT_OP_RD);
#if DEBUG_LEVEL > 5
	val = INW(dp, GMAC_REG(port, GM_SMI_CTRL));
	if ((val & GM_SMI_CT_OP_RD) == 0) {
		cmn_err(CE_WARN,
			"%s: %s: PHY read impossible on Port %d (Ctrl=0x%04x)",
			dp->name, __func__, port, val);
		return (0);
	}
#endif
	i = 0;
	while ((INW(dp, GMAC_REG(port, GM_SMI_CTRL)) & GM_SMI_CT_RD_VAL) == 0) {
		if (i++ > 10000) {
			cmn_err(CE_WARN, "%s: %s: timeout",
				dp->name, __func__);
			return (0);
		}
		drv_usecwait(10);
	}

	val = INW(dp, GMAC_REG(port, GM_SMI_DATA));

	DPRINTF(5, (CE_CONT, "%s: %s: reg:0x%x val:0x%x",
		dp->name, __func__, reg, val));

	return (val);
}


static void
myk_mii_write(struct gem_dev *dp, uint_t reg, uint16_t val)
{
	int		i;
	uint_t		port = PORT(dp);
	struct myk_dev	*lp = (struct myk_dev *) dp->private;

	DPRINTF(5, (CE_CONT, "!%s: %s: called, reg:0x%x, val:0x%x",
		dp->name, __func__, reg, val));

	OUTW(dp, GMAC_REG(port, GM_SMI_DATA), val);
	OUTW(dp, GMAC_REG(port, GM_SMI_CTRL),
		GM_SMI_CT_PHY_AD(dp->mii_phy_addr) | GM_SMI_CT_REG_AD(reg));

	i = 0;
	while ((INW(dp, GMAC_REG(port, GM_SMI_CTRL)) & GM_SMI_CT_BUSY) != 0) {
		if (i++ > 1000) {
			cmn_err(CE_WARN, "%s: %s: timeout",
				dp->name, __func__);
			break;
		}
		drv_usecwait(10);
	}
}

/*
 * myk_mii_config: phy initialization after phy reset
 */
static int
myk_mii_config(struct gem_dev *dp)
{
	uint16_t	pg;
	uint16_t	ctrl;
	uint16_t	ledctrl;
	uint16_t	ledover;
	struct myk_dev	*lp = (struct myk_dev *) dp->private;

	DPRINTF(0, (CE_CONT, "!%s: %s: called", dp->name, __func__));

	if  (gem_mii_config_default(dp) != GEM_SUCCESS) {
		return (GEM_FAILURE);
	}

	/* Setup Phy LED's */
	ledctrl = PHY_M_LED_PULS_DUR(PULS_170MS);
	ledover = 0;

	switch (lp->chip_id) {
	case CHIP_ID_YUKON_FE:
		/* on 88E3082 these bits are at 11..9 (shifted left) */
		ledctrl |= PHY_M_LED_BLINK_RT(BLINK_84MS) << 1;

		ctrl = myk_mii_read(dp, PHY_MARV_FE_LED_PAR);

		/* delete ACT LED control bits */
		ctrl &= ~PHY_M_FELP_LED1_MSK;
		/* change ACT LED control to blink mode */
		ctrl |= PHY_M_FELP_LED1_CTRL(LED_PAR_CTRL_ACT_BL);
		myk_mii_write(dp, PHY_MARV_FE_LED_PAR, ctrl);
		break;

	case CHIP_ID_YUKON_XL:
		pg = myk_mii_read(dp, PHY_MARV_EXT_ADR);

		/* select page 3 to access LED control register */
		myk_mii_write(dp, PHY_MARV_EXT_ADR, 3);

		/* set LED Function Control register */
		myk_mii_write(dp, PHY_MARV_PHY_CTRL,
			     (PHY_M_LEDC_LOS_CTRL(1) |	/* LINK/ACT */
			      PHY_M_LEDC_INIT_CTRL(7) |	/* 10 Mbps */
			      PHY_M_LEDC_STA1_CTRL(7) |	/* 100 Mbps */
			      PHY_M_LEDC_STA0_CTRL(7)));	/* 1000 Mbps */

		/* set Polarity Control register */
		myk_mii_write(dp, PHY_MARV_PHY_STAT,
			     (PHY_M_POLC_LS1_P_MIX(4) |
			      PHY_M_POLC_IS0_P_MIX(4) |
			      PHY_M_POLC_LOS_CTRL(2) |
			      PHY_M_POLC_INIT_CTRL(2) |
			      PHY_M_POLC_STA1_CTRL(2) |
			      PHY_M_POLC_STA0_CTRL(2)));

		/* restore page register */
		myk_mii_write(dp, PHY_MARV_EXT_ADR, pg);
		break;

	case CHIP_ID_YUKON_EC_U:
		pg = myk_mii_read(dp, PHY_MARV_EXT_ADR);

		/* select page 3 to access LED control register */
		myk_mii_write(dp, PHY_MARV_EXT_ADR, 3);

		/* set LED Function Control register */
		myk_mii_write(dp, PHY_MARV_PHY_CTRL,
			     (PHY_M_LEDC_LOS_CTRL(1) |	/* LINK/ACT */
			      PHY_M_LEDC_INIT_CTRL(8) |	/* 10 Mbps */
			      PHY_M_LEDC_STA1_CTRL(7) |	/* 100 Mbps */
			      PHY_M_LEDC_STA0_CTRL(7)));/* 1000 Mbps */

		/* set Blink Rate in LED Timer Control Register */
		myk_mii_write(dp, PHY_MARV_INT_MASK,
			     ledctrl | PHY_M_LED_BLINK_RT(BLINK_84MS));
		/* restore page register */
		myk_mii_write(dp, PHY_MARV_EXT_ADR, pg);
		break;

	default:
		/* set Tx LED (LED_TX) to blink mode on Rx OR Tx activity */
		ledctrl |= PHY_M_LED_BLINK_RT(BLINK_84MS) | PHY_M_LEDC_TX_CTRL;
		/* turn off the Rx LED (LED_RX) */
		ledover |= PHY_M_LED_MO_RX(MO_LED_OFF);
	}

	if (lp->chip_id == CHIP_ID_YUKON_EC_U &&
	    lp->chip_rev == CHIP_REV_YU_EC_A1) {
		/* apply fixes in PHY AFE */
		pg = myk_mii_read(dp, PHY_MARV_EXT_ADR);
		myk_mii_write(dp, PHY_MARV_EXT_ADR, 255);

		/* increase differential signal amplitude in 10BASE-T */
		myk_mii_write(dp, 0x18, 0xaa99);
		myk_mii_write(dp, 0x17, 0x2011);

		/* fix for IEEE A/B Symmetry failure in 1000BASE-T */
		myk_mii_write(dp, 0x18, 0xa204);
		myk_mii_write(dp, 0x17, 0x2002);

		/* set page register to 0 */
		myk_mii_write(dp, PHY_MARV_EXT_ADR, pg);
	} else {
		myk_mii_write(dp, PHY_MARV_LED_CTRL, ledctrl);

		if (!dp->mii_fixedmode || dp->speed == GEM_SPD_100) {
			/* turn on 100 Mbps LED (LED_LINK100) */
			ledover |= PHY_M_LED_MO_100(MO_LED_ON);
		}

		if (ledover) {
			myk_mii_write(dp, PHY_MARV_LED_OVER, ledover);
		}
	}

	/* Enable phy interrupt on auto-negotiation complete or link change  */
	myk_mii_write(dp, PHY_MARV_INT_MASK, (!dp->mii_fixedmode) ?
			(PHY_M_IS_AN_COMPL | PHY_M_DEF_MSK) : PHY_M_DEF_MSK);

	return (GEM_SUCCESS);
}

/*
 * myk_mii_init: phy initialization after power up
 */
static int
myk_mii_init(struct gem_dev *dp)
{
	uint16_t	pg;
	uint16_t	ctrl;
	struct myk_dev	*lp = (struct myk_dev *) dp->private;

	if ((!dp->mii_fixedmode) &&
	    !(lp->chip_id == CHIP_ID_YUKON_XL ||
	      lp->chip_id == CHIP_ID_YUKON_EC_U)) {
		uint16_t ectrl = myk_mii_read(dp, PHY_MARV_EXT_CTRL);

		ectrl &= ~(PHY_M_EC_M_DSC_MSK | PHY_M_EC_S_DSC_MSK |
			   PHY_M_EC_MAC_S_MSK);
		ectrl |= PHY_M_EC_MAC_S(MAC_TX_CLK_25_MHZ);

		if (lp->chip_id == CHIP_ID_YUKON_EC) {
			ectrl |= PHY_M_EC_DSC_2(2) | PHY_M_EC_DOWN_S_ENA;
		} else {
			ectrl |= PHY_M_EC_M_DSC(2) | PHY_M_EC_S_DSC(3);
		}
		myk_mii_write(dp, PHY_MARV_EXT_CTRL, ectrl);
	}

	ctrl = myk_mii_read(dp, PHY_MARV_PHY_CTRL);

	if (myk_is_copper(dp)) {
		if (lp->chip_id == CHIP_ID_YUKON_FE) {
			/* enable automatic crossover */
			ctrl |= PHY_M_PC_MDI_XMODE(PHY_M_PC_ENA_AUTO) >> 1;
		} else {
			/* disable energy detect */
			ctrl &= ~PHY_M_PC_EN_DET_MSK;

			/* enable automatic crossover */
			ctrl |= PHY_M_PC_MDI_XMODE(PHY_M_PC_ENA_AUTO);

			if (!dp->mii_fixedmode &&
			    (lp->chip_id == CHIP_ID_YUKON_XL ||
			     lp->chip_id == CHIP_ID_YUKON_EC_U)) {
				ctrl &= ~PHY_M_PC_DSC_MSK;
				ctrl |= PHY_M_PC_DSC(2) | PHY_M_PC_DOWN_S_ENA;
			}
		}
	} else {
		/* workaround for deviation #4.88 (CRC errors) */
		/* disable Automatic Crossover */

		ctrl &= ~PHY_M_PC_MDIX_MSK;
	}

	myk_mii_write(dp, PHY_MARV_PHY_CTRL, ctrl);

	/* special setup for PHY 88E1112 Fiber */
	if (lp->chip_id == CHIP_ID_YUKON_XL && !myk_is_copper(dp)) {
		pg = myk_mii_read(dp, PHY_MARV_EXT_ADR);

		/* Fiber: select 1000BASE-X only mode MAC Specific Ctrl Reg. */
		myk_mii_write(dp, PHY_MARV_EXT_ADR, 2);
		ctrl = myk_mii_read(dp, PHY_MARV_PHY_CTRL);
		ctrl &= ~PHY_M_MAC_MD_MSK;
		ctrl |= PHY_M_MAC_MODE_SEL(PHY_M_MAC_MD_1000BX);
		myk_mii_write(dp, PHY_MARV_PHY_CTRL, ctrl);

		if (lp->pmd_type == 'P') {
			/* select page 1 to access Fiber registers */
			myk_mii_write(dp, PHY_MARV_EXT_ADR, 1);

			/* for SFP-module set SIGDET polarity to low */
			ctrl = myk_mii_read(dp, PHY_MARV_PHY_CTRL);
			ctrl |= PHY_M_FIB_SIGD_POL;
			myk_mii_write(dp, PHY_MARV_CTRL, ctrl);
		}

		myk_mii_write(dp, PHY_MARV_EXT_ADR, pg);
	}

	return (GEM_SUCCESS);
}


/* ======================================================== */
/*
 * OS depend (device driver) routine
 */
/* ======================================================== */
static int
mykattach(dev_info_t *dip, ddi_attach_cmd_t cmd)
{
	int			i;
	int			t8;
	int			n;
	ddi_acc_handle_t	conf_handle;
	int			vid;
	int			did;
	uint8_t			revid;
	uint8_t			lat;
	uint8_t			chip_id;
	uint8_t			chip_rev;
	uint8_t			pm_cap;
	uint8_t			pmd_type;
	int			ports;
	int			unit;
	boolean_t		is_pcie;
	const char		*drv_name;
	struct gem_dev		*dp[2];
	struct myk_dev		*lp;
	caddr_t			base;
	ddi_acc_handle_t	regs_ha;
	struct gem_conf		*gcp;

	unit =  ddi_get_instance(dip);
	drv_name = ddi_driver_name(dip);

	DPRINTF(0, (CE_CONT, "%s%d: mykattach: called %s",
		drv_name, unit, drv_name));

	/*
	 * Check if chip is supported.
	 */
	if (pci_config_setup(dip, &conf_handle) != DDI_SUCCESS) {
		cmn_err(CE_WARN, "%s: ddi_regs_map_setup failed",
			drv_name);
		goto err;
	}

	vid  = pci_config_get16(conf_handle, PCI_CONF_VENID);
	did  = pci_config_get16(conf_handle, PCI_CONF_DEVID);
	revid= pci_config_get8(conf_handle, PCI_CONF_REVID);
	lat  = pci_config_get8(conf_handle, PCI_CONF_LATENCY_TIMER);
#if 0
	for (i = 0, p = myk_chiptbl; i < CHIPTABLESIZE; i++, p++) {
		if (p->venid == vid && p->devid == did) {
			/* found */
			goto chip_found;
		}
	}

	/* Not found */
	cmn_err(CE_WARN,
		"%s: myk_attach: wrong PCI venid/devid (0x%x, 0x%x)",
		drv_name, vid, did);
	pci_config_teardown(&conf_handle);
	goto err;
#endif
chip_found:
	/* ensure the pmr status is D0 mode */
	(void) gem_pci_set_power_state(dip, conf_handle, PCI_PMCSR_D0);

	pci_config_put16(conf_handle, PCI_CONF_COMM,
		PCI_COMM_IO | PCI_COMM_MAE | PCI_COMM_ME |
			pci_config_get16(conf_handle, PCI_CONF_COMM));


	cmn_err(CE_CONT,
"!%s%d: (vid: 0x%04x, did: 0x%04x, revid: 0x%02x, latency timer: 0x%02x)",
		drv_name, unit, vid, did, revid, lat);

	switch (cmd) {
	case DDI_RESUME:
		pci_config_teardown(&conf_handle);
		return (gem_resume(dip));

	case DDI_ATTACH:
		if (gem_pci_regs_map_setup(dip,
			PCI_ADDR_MEM64, PCI_ADDR_MASK,
			&myk_dev_attr, (caddr_t *)&base, &regs_ha)
					!= DDI_SUCCESS) {
			goto err;
		}

		/*
		 * Global reset and check configuration
		 */
		ddi_put8(regs_ha, (uint8_t *)(base + B0_CTST), CS_RST_CLR);

		chip_id = ddi_get8(regs_ha, (uint8_t *)(base + B2_CHIP_ID));
		if (chip_id < CHIP_ID_YUKON_XL || chip_id > CHIP_ID_YUKON_FE) {
			cmn_err(CE_WARN,
				"%s%d: unsupported chip type 0x%x\n",
				drv_name, unit, chip_id);
			ddi_regs_map_free(&regs_ha);
			goto err;
		}

		chip_rev = (ddi_get8(regs_ha, (uint8_t *)(base + B2_MAC_CFG))
				& CFG_CHIP_R_MSK) >> 4;

		/* This rev is really old, and requires untested workarounds */
		if (chip_id == CHIP_ID_YUKON_EC &&
		    chip_rev == CHIP_REV_YU_EC_A1) {
			cmn_err(CE_WARN,
			"%s%d: unsupported revision Yukon-%s (0x%x) rev %d\n",
				drv_name, unit,
				yukon2_name[chip_id - CHIP_ID_YUKON_XL],
				chip_id, chip_rev);
			ddi_regs_map_free(&regs_ha);
			goto err;
		}

		/* disable ASF */
		if (chip_id <= CHIP_ID_YUKON_EC) {
			ddi_put8(regs_ha,
				(uint8_t *) (base + B28_Y2_ASF_STAT_CMD),
				Y2_ASF_RESET);
			ddi_put16(regs_ha,
				(uint16_t *)(base + B0_CTST), Y2_ASF_DISABLE);
		}

		ddi_put16(regs_ha, (uint16_t *)(base + B0_CTST), CS_RST_CLR);

		/* Disable ASF. */
		if (chip_id < CHIP_ID_YUKON_XL) {
			ddi_put32(regs_ha,
				(uint32_t *) (base + B28_Y2_ASF_STAT_CMD),
				Y2_ASF_RESET); /* b */
			ddi_put16(regs_ha, 
				(uint16_t *) (base + B0_CTST),
				Y2_ASF_DISABLE); /* b */
		}

		/*
		 * Since we disabled ASF, S/W reset is required for
		 * Power Management.
		 */
		ddi_put16(regs_ha, (uint16_t *) (base + B0_CTST), CS_RST_SET);
		ddi_put16(regs_ha, (uint16_t *) (base + B0_CTST), CS_RST_CLR);

		pmd_type = ddi_get8(regs_ha, (uint8_t *) (base + B2_PMD_TYP));
		ports = 1;

		t8 = ddi_get8(regs_ha, (uint8_t *)(base + B2_Y2_HW_RES));
		if ((t8 & CFG_DUAL_MAC_MSK) == CFG_DUAL_MAC_MSK) {
			if (!(ddi_get8(regs_ha,
				(uint8_t *) (base + B2_Y2_CLK_GATE))
						& Y2_STATUS_LNK2_INAC)) {
				ports++;
			}
		}

		is_pcie = gem_is_pcie(dip, conf_handle);
		pm_cap = gem_search_pci_cap(dip, conf_handle, PCI_CAP_ID_PM);

		cmn_err(CE_NOTE,
		"%s%d: Yukon-%s (0x%x) rev %d, pcie:%d pm_cap:0x%x nport:%d",
			drv_name, unit,
		(chip_id >= CHIP_ID_YUKON_XL && chip_id <= CHIP_ID_YUKON_FE) 
				? yukon2_name[chip_id - CHIP_ID_YUKON_XL]
				: "UNKNOWN",
			chip_id, chip_rev, is_pcie, pm_cap, ports);

		pci_config_teardown(&conf_handle);
		conf_handle = NULL;

		/*
		 * construct gem configration
		 */
		gcp = (struct gem_conf *) kmem_zalloc(sizeof(*gcp), KM_SLEEP);

		/* name */
#ifdef notdef
		sprintf(gcp->gc_name, "%s%d", drv_name, unit);
#else
		sprintf(gcp->gc_name, "%s%%d", drv_name);
#endif
		/* configuration of tx and rx rings and buffers */
		gcp->gc_tx_buf_align = sizeof(uint8_t) - 1;
		gcp->gc_tx_max_frags = GEM_MAXTXFRAGS;
		gcp->gc_tx_max_descs_per_pkt = gcp->gc_tx_max_frags*2 + 1;
		gcp->gc_tx_desc_unit_shift = 3;		/* 8 byte */
		gcp->gc_tx_buf_size  = TX_BUF_SIZE;
		gcp->gc_tx_buf_limit = gcp->gc_tx_buf_size;
		gcp->gc_tx_ring_size = TX_RING_SIZE;
		gcp->gc_tx_ring_limit = gcp->gc_tx_ring_size;
		gcp->gc_tx_auto_pad  = B_TRUE;
		gcp->gc_tx_copy_thresh = myk_tx_copy_thresh;
		gcp->gc_tx_desc_write_oo =
			(gcp->gc_tx_max_descs_per_pkt == 1) &&
			(TX_RING_SIZE == TX_BUF_SIZE);

		gcp->gc_rx_buf_align = sizeof(uint32_t) - 1;
		gcp->gc_rx_max_frags = 1;
		/* XXX - we allocate rx descriptors in io_area by ourselves */
		gcp->gc_rx_desc_unit_shift = -1;
		gcp->gc_rx_ring_size = RX_RING_SIZE;
		gcp->gc_rx_buf_max   = min(RX_BUF_SIZE, (RX_RING_SIZE/2)-2);
		gcp->gc_rx_copy_thresh = myk_rx_copy_thresh;

		gcp->gc_io_area_size = RX_RING_BYTES + STATUS_LE_BYTES;

		/* map attributes */
		gcp->gc_dev_attr = myk_dev_attr;
		gcp->gc_buf_attr = myk_buf_attr;
		gcp->gc_desc_attr = myk_buf_attr;

		/* dma attributes */
		gcp->gc_dma_attr_desc = myk_dma_attr_desc;

		gcp->gc_dma_attr_txbuf = myk_dma_attr_buf;
		gcp->gc_dma_attr_txbuf.dma_attr_align = gcp->gc_tx_buf_align+1;
		gcp->gc_dma_attr_txbuf.dma_attr_sgllen = gcp->gc_tx_max_frags;

		gcp->gc_dma_attr_rxbuf = myk_dma_attr_buf;
		gcp->gc_dma_attr_rxbuf.dma_attr_align = gcp->gc_rx_buf_align+1;
		gcp->gc_dma_attr_rxbuf.dma_attr_sgllen = gcp->gc_rx_max_frags;

		/* time out parameters */
		gcp->gc_tx_timeout = 5*ONESEC;
		gcp->gc_tx_timeout_interval = ONESEC;

		/* flow control */
		gcp->gc_flow_control = FLOW_CONTROL_RX_PAUSE;

		/* mii mode */
		gcp->gc_mii_mode = GEM_MODE_1000BASET;

		/* MII timeout parameters */
		gcp->gc_mii_link_watch_interval = ONESEC;
		gcp->gc_mii_an_watch_interval   = ONESEC/5;
		gcp->gc_mii_reset_timeout = MII_RESET_TIMEOUT;	/* 1 sec */
		gcp->gc_mii_an_timeout    = MII_AN_TIMEOUT;	/* 5 sec */
		gcp->gc_mii_an_wait       = 0;
		gcp->gc_mii_linkdown_timeout = MII_LINKDOWN_TIMEOUT;
		gcp->gc_mii_hw_link_detection = B_TRUE;

		/* MII work arounds */
		gcp->gc_mii_addr_min = 0;
		gcp->gc_mii_an_delay = 0;
#if 0
		gcp->gc_mii_linkdown_action = MII_ACTION_RSA;
#else
		gcp->gc_mii_linkdown_action = MII_ACTION_RESET;
#endif
		gcp->gc_mii_linkdown_timeout_action = MII_ACTION_RESET;
		gcp->gc_mii_dont_reset = B_FALSE;

		/* I/O methods */

		/* mac operation */
		gcp->gc_attach_chip = &myk_attach_port;
		gcp->gc_reset_chip  = &myk_reset_port;
		gcp->gc_init_chip   = &myk_init_port;
		gcp->gc_start_chip  = &myk_start_port;
		gcp->gc_stop_chip   = &myk_stop_port;
		gcp->gc_multicast_hash = &myk_mcast_hash;
		gcp->gc_set_rx_filter = &myk_set_rx_filter;
		gcp->gc_set_media   = &myk_set_media;
		gcp->gc_get_stats   = &myk_get_stats;
		gcp->gc_interrupt   = &myk_soft_intr;

		/* descriptor operation */
		gcp->gc_tx_desc_write = &myk_tx_desc_write;
		gcp->gc_rx_desc_write = &myk_rx_desc_write;
		gcp->gc_tx_start      = &myk_tx_start;
		gcp->gc_rx_start      = &myk_rx_start;

		gcp->gc_tx_desc_init = &myk_tx_desc_init;
		gcp->gc_rx_desc_init = &myk_rx_desc_init;
		gcp->gc_tx_desc_stat = &myk_tx_desc_stat;
		gcp->gc_rx_desc_stat = &myk_rx_desc_stat;
		gcp->gc_tx_desc_clean = &myk_tx_desc_init;
		gcp->gc_rx_desc_clean = &myk_rx_desc_init;

		/* mii operations */
		gcp->gc_mii_probe  = &gem_mii_probe_default;
		gcp->gc_mii_init = &myk_mii_init;
		gcp->gc_mii_config = &myk_mii_config;
		gcp->gc_mii_sync  = &myk_mii_sync;
		gcp->gc_mii_read  = &myk_mii_read;
		gcp->gc_mii_write = &myk_mii_write;
		gcp->gc_mii_tune_phy = NULL;

		/* number of max ports in specification */
		gcp->gc_nports = 2;

		lp = kmem_zalloc(sizeof(struct myk_dev), KM_SLEEP);
		lp->chip_id = chip_id;
		lp->chip_id = chip_rev;
		lp->is_pcie = is_pcie;
		lp->pm_cap  = pm_cap;
		lp->pmd_type = pmd_type;
		lp->ports   = ports;

		/* initialize head of port list */
		ddi_set_driver_private(dip, NULL);
		dp[0] = dp[1] = NULL;

		for (i = 0; i < ports; i++) {
			dp[i] = gem_do_attach(dip, i,
				gcp, base, &regs_ha, lp, sizeof(*lp));

			if (dp[i] == NULL) {
				goto err_free_gc;
			}
		}

		kmem_free(gcp, sizeof(*gcp));

		if (dp[0]->misc_flag & GEM_SOFTINTR) {
			/* add real hw interrupt handler */
			if (ddi_add_intr(dip, 0, &lp->iblock_cookie, NULL,
				(u_int (*)(caddr_t))myk_interrupt,
				(caddr_t)dp[0]) != DDI_SUCCESS) {
				cmn_err(CE_WARN, "!%s%d: ddi_add_intr failed",
					drv_name, unit);
				goto err_free_mem;
			}
		}

		myk_status_desc_init(lp);

		return (DDI_SUCCESS);

err_free_gc:
		kmem_free(gcp, sizeof(*gcp));
err_free_mem:
		kmem_free(lp, sizeof(struct myk_dev));
err:
		if (conf_handle != NULL) {
			pci_config_teardown(&conf_handle);
		}

		return (DDI_FAILURE);
	}
	return (DDI_FAILURE);
}

static int
mykdetach(dev_info_t *dip, ddi_detach_cmd_t cmd)
{
	struct gem_dev  *dp;
	struct myk_dev	*lp;
	int	ret;

	dp = ddi_get_driver_private(dip);
	lp = (struct myk_dev *) dp->private;

	/* turn off LED's */
	OUTW(dp, B0_CTST, Y2_LED_STAT_OFF);

	/* stop the hardware completely */
	myk_set_power_state(dp, PCI_PMCSR_D3HOT);
	OUTW(dp, B0_CTST, Y2_LED_STAT_OFF);
	OUTB(dp, B0_CTST, CS_RST_SET);
	FLSHB(dp, B0_CTST);

	switch (cmd) {
	case DDI_SUSPEND:
		return  (gem_suspend(dip));

	case DDI_DETACH:
		if (dp->misc_flag & GEM_SOFTINTR) {
			ddi_remove_intr(dip, 0, lp->iblock_cookie);
		}
		gem_do_detach(dip);
		return (DDI_SUCCESS);
	}
	return (DDI_FAILURE);
}

/* ======================================================== */
/*
 * OS depend (loadable streams driver) routine
 */
/* ======================================================== */
#ifdef GEM_CONFIG_GLDv3
DDI_DEFINE_STREAM_OPS(myk_ops, nulldev, nulldev, mykattach, mykdetach,
nodev, NULL, D_MP, NULL);
#else
static	struct module_info mykminfo = {
	0,			/* mi_idnum */
	"myk",			/* mi_idname */
	0,			/* mi_minpsz */
	INFPSZ,			/* mi_maxpsz */
	64*1024,		/* mi_hiwat */
	1,			/* mi_lowat */
};

static	struct qinit mykrinit = {
	(int (*)()) NULL,	/* qi_putp */
	gem_rsrv,		/* qi_srvp */
	gem_open,		/* qi_qopen */
	gem_close,		/* qi_qclose */
	(int (*)()) NULL,	/* qi_qadmin */
	&mykminfo,		/* qi_minfo */
	NULL			/* qi_mstat */
};

static	struct qinit mykwinit = {
	gem_wput,		/* qi_putp */
	gem_wsrv,		/* qi_srvp */
	(int (*)()) NULL,	/* qi_qopen */
	(int (*)()) NULL,	/* qi_qclose */
	(int (*)()) NULL,	/* qi_qadmin */
	&mykminfo,		/* qi_minfo */
	NULL			/* qi_mstat */
};
static struct streamtab	myk_info = {
	&mykrinit,	/* st_rdinit */
	&mykwinit,	/* st_wrinit */
	NULL,		/* st_muxrinit */
	NULL		/* st_muxwrinit */
};

static	struct cb_ops cb_myk_ops = {
	nulldev,	/* cb_open */
	nulldev,	/* cb_close */
	nodev,		/* cb_strategy */
	nodev,		/* cb_print */
	nodev,		/* cb_dump */
	nodev,		/* cb_read */
	nodev,		/* cb_write */
	nodev,		/* cb_ioctl */
	nodev,		/* cb_devmap */
	nodev,		/* cb_mmap */
	nodev,		/* cb_segmap */
	nochpoll,	/* cb_chpoll */
	ddi_prop_op,	/* cb_prop_op */
	&myk_info,	/* cb_str */
	D_MP,		/* cb_flag */
#ifdef notdef
	CB_REV,		/* cb_rev */
	nodev,		/* cb_aread */
	nodev,		/* cb_awrite */
#endif
};

static	struct dev_ops myk_ops = {
	DEVO_REV,	/* devo_rev */
	0,		/* devo_refcnt */
	gem_getinfo,	/* devo_getinfo */
	nulldev,	/* devo_identify */
	nulldev,	/* devo_probe */
	mykattach,	/* devo_attach */
	mykdetach,	/* devo_detach */
	nodev,		/* devo_reset */
	&cb_myk_ops,	/* devo_cb_ops */
	NULL,		/* devo_bus_ops */
	gem_power	/* devo_power */
};
#endif /* GEM_CONFIG_GLDv3 */

static struct modldrv modldrv = {
	&mod_driverops,	/* Type of module.  This one is a driver */
	ident,
	&myk_ops,	/* driver ops */
};

static struct modlinkage modlinkage = {
	MODREV_1, &modldrv, NULL
};

/* ======================================================== */
/*
 * _init : done
 */
/* ======================================================== */
int
_init(void)
{
	int 	status;

	DPRINTF(2, (CE_CONT, "myk: _init: called"));
	gem_mod_init(&myk_ops, "myk");
	status = mod_install(&modlinkage);
	if (status != DDI_SUCCESS) {
		gem_mod_fini(&myk_ops);
	}

	return (status);
}

/*
 * _fini : done
 */
int
_fini(void)
{
	int	status;

	DPRINTF(2, (CE_CONT, "myk: _fini: called"));
	status = mod_remove(&modlinkage);
	if (status == DDI_SUCCESS) {
		gem_mod_fini(&myk_ops);
	}
	return (status);
}

int
_info(struct modinfo *modinfop)
{
	return (mod_info(&modlinkage, modinfop));
}
